# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import functools

import mock
import testscenarios
import testtools
from webob import exc

from oslo_config import cfg

from neutron_lib.api.definitions import external_net
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory

from neutron.tests.unit import _test_extension_portbindings as test_bindings
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.db import test_allowedaddresspairs_db as test_addr
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_plugin
from neutron.tests.unit.extensions import test_extra_dhcp_opt as test_dhcpopts
from neutron.tests.unit.extensions import test_extraroute as test_ext_route
from neutron.tests.unit.extensions import test_l3
from neutron.tests.unit.extensions import test_l3_ext_gw_mode as test_gw_mode
from neutron.tests.unit.extensions import test_portsecurity as test_psec
from neutron.tests.unit.extensions import test_securitygroup as test_sg

from midonet.neutron.common import constants as m_const
from midonet.neutron.tests.unit import test_midonet_plugin as test_mn_plugin

load_tests = testscenarios.load_tests_apply_scenarios

PLUGIN_NAME = 'neutron.plugins.ml2.plugin.Ml2Plugin'

cfg.CONF.import_group('ml2', 'neutron.conf.plugins.ml2.config')


class MidonetPluginConf(object):
    """Plugin configuration shared across the unit and functional tests."""

    plugin_name = PLUGIN_NAME

    @staticmethod
    def setUp(test_case, parent_setup=None):
        """Perform additional configuration around the parent's setUp."""
        ovo_push_interface_p = mock.patch(
            'neutron.plugins.ml2.ovo_rpc.OVOServerRpcInterface')
        ovo_push_interface_p.start()
        cfg.CONF.set_override('client', test_mn_plugin.TEST_MN_CLIENT,
                              group='MIDONET')
        cfg.CONF.set_override('type_drivers',
                              ['midonet', 'uplink', 'local'],
                              group='ml2')
        cfg.CONF.set_override('mechanism_drivers',
                              ['midonet', 'openvswitch'],
                              group='ml2')
        cfg.CONF.set_override('tenant_network_types',
                              ['midonet'],
                              group='ml2')
        cfg.CONF.set_override('extension_drivers',
                              ['port_security'],
                              group='ml2')
        if parent_setup:
            parent_setup()


class MidonetPluginML2TestCase(test_plugin.NeutronDbPluginV2TestCase):

    def setup_parent(self, service_plugins=None, ext_mgr=None):

        # Set up mock for the midonet client to be made available in tests
        patcher = mock.patch(test_mn_plugin.TEST_MN_CLIENT)
        self.client_mock = mock.Mock()
        patcher.start().return_value = self.client_mock

        l3_plugin = {'l3_plugin_name': 'midonet_l3'}

        if service_plugins:
            service_plugins.update(l3_plugin)
        else:
            service_plugins = l3_plugin

        # Ensure that the parent setup can be called without arguments
        # by the common configuration setUp.
        parent_setup = functools.partial(
            super(MidonetPluginML2TestCase, self).setUp,
            plugin=MidonetPluginConf.plugin_name,
            service_plugins=service_plugins,
            ext_mgr=ext_mgr,
        )
        MidonetPluginConf.setUp(self, parent_setup)
        self.port_create_status = 'DOWN'

    def setUp(self, plugin=None, service_plugins=None, ext_mgr=None):
        self.setup_parent(service_plugins=service_plugins, ext_mgr=ext_mgr)


class TestMidonetNetworksML2(MidonetPluginML2TestCase,
                             test_plugin.TestNetworksV2):
    pass


class TestMidonetSecurityGroupML2(test_sg.TestSecurityGroups,
                                  MidonetPluginML2TestCase):
    pass


class TestMidonetSubnetsML2(MidonetPluginML2TestCase,
                            test_plugin.TestSubnetsV2):

    def test_delete_subnet_port_exists_owned_by_network(self):
        # MidoNet doesn't support updating dhcp port's fixed-ips.
        pass

    def test_delete_subnet_dhcp_port_associated_with_other_subnets(self):
        # MidoNet doesn't support updating dhcp port's fixed-ips.
        pass


class TestMidonetPortsML2(MidonetPluginML2TestCase,
                          test_plugin.TestPortsV2):

    def test_update_dhcp_port_with_exceeding_fixed_ips(self):
        # MidoNet doesn't support updating dhcp port's fixed-ips.
        pass


class TestMidonetRouterML2(MidonetPluginML2TestCase,
                           test_l3.L3NatTestCaseMixin):
    scenarios = [
        ('midonet',
         dict(network_type='midonet',
              expected_code=exc.HTTPOk.code,
              expected_code_for_create=exc.HTTPCreated.code)),
        ('uplink',
         dict(network_type='uplink',
              expected_code=exc.HTTPOk.code,
              expected_code_for_create=exc.HTTPCreated.code)),
        ('local',
         dict(network_type='local',
              expected_code=exc.HTTPBadRequest.code,
              expected_code_for_create=exc.HTTPBadRequest.code)),
    ]

    def setUp(self):
        ext_mgr = test_l3.L3TestExtensionManager()
        super(TestMidonetRouterML2, self).setUp(ext_mgr=ext_mgr)
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)

    def test_router_gateway_incompatible_network_for_create(self):
        if self.network_type == 'uplink':
            self.skipTest('This test case is not appropriate for uplink')
        net_param = {
            pnet.NETWORK_TYPE: self.network_type,
            external_net.EXTERNAL: True,
            'arg_list': (pnet.NETWORK_TYPE, external_net.EXTERNAL),
        }
        with self.network(**net_param) as net, \
                self.subnet(network=net):
            gw_info = {
                'network_id': net['network']['id'],
            }
            res = self._create_router(self.fmt, tenant_id=None,
                                      arg_list=('external_gateway_info',),
                                      external_gateway_info=gw_info)
            self.assertEqual(self.expected_code_for_create, res.status_int)

    def test_router_gateway_incompatible_network_for_update(self):
        if self.network_type == 'uplink':
            self.skipTest('This test case is not appropriate for uplink')
        net_param = {
            pnet.NETWORK_TYPE: self.network_type,
            external_net.EXTERNAL: True,
            'arg_list': (pnet.NETWORK_TYPE, external_net.EXTERNAL),
        }
        with self.network(**net_param) as net, \
                self.subnet(network=net), \
                self.router() as r:
            self._add_external_gateway_to_router(
                router_id=r['router']['id'],
                network_id=net['network']['id'],
                expected_code=self.expected_code)

    def test_router_interface_incompatible_network(self):
        net_param = {
            pnet.NETWORK_TYPE: self.network_type,
            'arg_list': (pnet.NETWORK_TYPE,),
        }
        with self.network(**net_param) as net, \
                self.subnet(network=net) as subnet, \
                self.router() as r:
            self._router_interface_action(
                'add', r['router']['id'], subnet['subnet']['id'], None,
                expected_code=self.expected_code)


class TestMidonetL3NatExtraRoute(test_ext_route.ExtraRouteDBIntTestCase,
                                 MidonetPluginML2TestCase):

    def test_router_update_gateway_upon_subnet_create_max_ips_ipv6(self):
        # MidoNet doesn't support IPv6.
        # MidoNet doesn't support fixed_ips updates on a router gateway port.
        pass

    def test_router_remove_ipv6_subnet_from_interface(self):
        # MidoNet doesn't support IPv6.
        # This specific case examines _add_interface_by_subnet,
        # which ends up with updating the existing router interface port's
        # fixed-ips.
        pass

    def test_router_add_interface_multiple_ipv6_subnets_same_net(self):
        # MidoNet doesn't support IPv6.
        # This specific case examines _add_interface_by_subnet,
        # which ends up with updating the existing router interface port's
        # fixed-ips.
        pass

    def test_router_update_gateway_add_multiple_prefixes_ipv6(self):
        # MidoNet doesn't support IPv6.
        # This specific case examines updating a router's ext_ips,
        # which ends up with updating its gateway port's fixed-ips.
        pass

    def test_router_update_gateway_upon_subnet_create_ipv6(self):
        # MidoNet doesn't support IPv6.
        # Even for IPv4, with the reference implementation, create_subnet
        # can ends up with adding an IP to the router's gateway port.
        # However, it can't happen for us because we reject a gateway port
        # without IP addresses.  ("No IPs assigned to the gateway port")
        pass

    def test_router_update_gateway_with_different_external_subnet(self):
        # This specific case examines updating a router's ext_ips,
        # which ends up with updating its gateway port's fixed-ips.
        pass

    def test_router_add_gateway_no_subnet(self):
        # Midonet does not support the case where a gateway is set
        # without a subnet, therefore we don't want to test this.
        pass

    def test_create_floatingip_ipv6_only_network_returns_400(self):
        # MidoNet supports floating IPv6, where it is possible to assign a
        # floating IP on an external network with only an IPv6 subnet.
        pass

    def test_create_router_no_gateway_ip(self):
        with self.network() as net:
            self._set_net_external(net['network']['id'])
            router_data = {'router': {
                'tenant_id': 'tenant_one',
                'external_gateway_info': {
                    'network_id': net['network']['id']}}}
            router_req = self.new_create_request('routers', router_data,
                                                 self.fmt)
            res = router_req.get_response(self.ext_api)
            self.assertEqual(400, res.status_int)

    def test_add_router_interface_by_port_failure(self):
        class _MyException(Exception):
            pass

        with self.port() as port, self.router() as router:
            ctx = context.get_admin_context()
            plugin = directory.get_plugin()
            l3_plugin = directory.get_plugin(plugin_constants.L3)
            router_id = router['router']['id']
            port_id = port['port']['id']
            interface_info = {
                'port_id': port_id,
            }
            with mock.patch.object(l3_plugin.client,
                                   'add_router_interface_postcommit',
                                   auto_spec=True,
                                   side_effect=_MyException), \
                    testtools.ExpectedException(_MyException):
                l3_plugin.add_router_interface(ctx, router_id, interface_info)
            port2 = plugin.get_port(ctx, port_id)
            self.assertEqual(port_id, port2['id'])

    def test_update_floatingip_error_change_resource_status_to_error(self):
        self.client_mock.update_floatingip_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.port() as p:
            private_sub = {
                'subnet': {
                    'id': p['port']['fixed_ips'][0]['subnet_id']
                }
            }
            with self.floatingip_no_assoc(private_sub) as fip:
                data = {'floatingip': {'port_id': p['port']['id']}}
                req = self.new_update_request('floatingips',
                                              data,
                                              fip['floatingip']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(exc.HTTPInternalServerError.code,
                                 res.status_int)
                req = self.new_show_request(
                    'floatingips', fip['floatingip']['id'])
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                self.assertEqual(
                    n_const.FLOATINGIP_STATUS_ERROR,
                    res['floatingip']['status'])

    def test_update_router_error_change_resource_status_to_error(self):
        self.client_mock.update_router_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.subnet(cidr='11.0.0.0/24') as pub_sub:
            pub_net = pub_sub['subnet']['network_id']
            self._set_net_external(pub_net)
            with self.router() as r:
                data = {'router':
                        {'external_gateway_info': {'network_id': pub_net}}}
                req = self.new_update_request('routers',
                                              data,
                                              r['router']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(exc.HTTPInternalServerError.code,
                                 res.status_int)
                req = self.new_show_request(
                    'routers', r['router']['id'])
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                self.assertEqual('ERROR', res['router']['status'])

    def test_floatingip_via_router_interface_returns_404(self):
        self.skipTest('Not appropriate with router-interface-fip extension.')

    def test_floatingip_via_router_interface_returns_201(self):
        self._test_floatingip_via_router_interface(exc.HTTPCreated.code)

    # NOTE(yamamoto): ML2 no longer uses NeutronDbPluginV2.create_port
    def test_floatingip_with_invalid_create_port(self):
        self._test_floatingip_with_invalid_create_port(PLUGIN_NAME)


class TestMidonetProviderNet(MidonetPluginML2TestCase):

    @contextlib.contextmanager
    def provider_net(self, name='name1', net_type=m_const.TYPE_UPLINK,
                     admin_state_up=True):
        args = {pnet.NETWORK_TYPE: net_type,
                'tenant_id': 'admin'}
        net = self._make_network(self.fmt, name, admin_state_up,
                                 arg_list=(pnet.NETWORK_TYPE, 'tenant_id'),
                                 **args)
        yield net

    def test_create_provider_net(self):
        keys = {pnet.NETWORK_TYPE: m_const.TYPE_UPLINK,
                'name': 'name1'}
        with self.provider_net() as net:
            self.assertDictSupersetOf(keys, net['network'])

    def test_create_provider_net_with_bogus_type(self):
        # Create with a bogus network type
        with testtools.ExpectedException(exc.HTTPClientError), \
                self.provider_net(net_type="random"):
            pass

    def test_create_provider_net_with_local(self):
        with self.provider_net(net_type=n_const.TYPE_LOCAL) as net:
            self.assertEqual(n_const.TYPE_LOCAL,
                             net['network'][pnet.NETWORK_TYPE])

    def test_create_provider_net_with_flat(self):
        with testtools.ExpectedException(exc.HTTPClientError), \
                self.provider_net(net_type=n_const.TYPE_FLAT):
            pass

    def test_create_provider_net_with_gre(self):
        with testtools.ExpectedException(exc.HTTPClientError), \
                self.provider_net(net_type=n_const.TYPE_GRE):
            pass

    def test_create_provider_net_with_vlan(self):
        with testtools.ExpectedException(exc.HTTPClientError), \
                self.provider_net(net_type=n_const.TYPE_VLAN):
            pass

    def test_create_provider_net_with_vxlan(self):
        with testtools.ExpectedException(exc.HTTPClientError), \
                self.provider_net(net_type=n_const.TYPE_VXLAN):
            pass

    def test_create_provider_net_with_geneve(self):
        with testtools.ExpectedException(exc.HTTPClientError), \
                self.provider_net(net_type=n_const.TYPE_GENEVE):
            pass

    def test_create_provider_net_with_midonet(self):
        with self.provider_net(net_type=m_const.TYPE_MIDONET) as net:
            self.assertEqual(m_const.TYPE_MIDONET,
                             net['network'][pnet.NETWORK_TYPE])

    def test_create_provider_net_without_type(self):
        args = {'network': {'tenant_id': 'admin'}}
        req = self.new_create_request('networks', args, self.fmt)
        res = req.get_response(self.api)
        self.assertEqual(201, res.status_int)
        net_res = self.deserialize(self.fmt, res)
        self.assertEqual(m_const.TYPE_MIDONET,
                         net_res['network'][pnet.NETWORK_TYPE])

    def test_update_provider_net_unsupported(self):
        # Update including the network type is not supported
        with self.provider_net() as net:
            args = {"network": {"name": "foo",
                                pnet.NETWORK_TYPE: m_const.TYPE_UPLINK}}
            req = self.new_update_request('networks', args,
                                          net['network']['id'])
            res = req.get_response(self.api)
            self.assertEqual(400, res.status_int)

    def test_delete_provider_net(self):
        with self.provider_net() as net:
            req = self.new_delete_request('networks', net['network']['id'])
            res = req.get_response(self.api)
            self.assertEqual(exc.HTTPNoContent.code, res.status_int)

    def test_show_provider_net(self):
        with self.provider_net() as net:
            req = self.new_show_request('networks', net['network']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertEqual(m_const.TYPE_UPLINK,
                             res['network'][pnet.NETWORK_TYPE])

    def test_list_provider_nets(self):
        # Create two uplink prov nets and retrieve them
        with self.provider_net():
            with self.provider_net(name="net2"):
                req = self.new_list_request('networks')
                res = self.deserialize(
                    self.fmt, req.get_response(self.api))
                self.assertEqual(2, len(res['networks']))
                for res_net in res['networks']:
                    self.assertEqual(m_const.TYPE_UPLINK,
                                     res_net[pnet.NETWORK_TYPE])

    def test_list_provider_nets_filtered_by_invalid_type(self):
        # Search a list of two provider networks with type uplink and type vlan
        with self.provider_net(name="net2"):
            with self.provider_net(name="net2"):
                params_str = "%s=%s" % (pnet.NETWORK_TYPE, 'vlan')
                req = self.new_list_request('networks', None,
                                            params=params_str)
                res = self.deserialize(
                    self.fmt, req.get_response(self.api))
                self.assertEqual(0, len(res['networks']))


class TestMidonetPortBinding(MidonetPluginML2TestCase,
                             test_bindings.PortBindingsTestCase):

    # Test case does not set binding:host_id, so ml2 does not attempt
    # to bind port
    VIF_TYPE = portbindings.VIF_TYPE_UNBOUND
    HAS_PORT_FILTER = True

    @contextlib.contextmanager
    def port_with_binding_profile(self, host='host', if_name='if_name'):
        args = {portbindings.PROFILE: {'interface_name': if_name},
                portbindings.HOST_ID: host}
        with test_plugin.optional_ctx(None, self.subnet) as subnet_to_use:
            net_id = subnet_to_use['subnet']['network_id']
            port = self._make_port(self.fmt, net_id,
                                   arg_list=(portbindings.PROFILE,
                                             portbindings.HOST_ID,), **args)
            yield port

    def test_create_mido_portbinding(self):
        keys = {portbindings.PROFILE: {'interface_name': 'if_name'},
                portbindings.HOST_ID: 'host'}
        with self.port_with_binding_profile() as port:
            self.assertDictSupersetOf(keys, port['port'])

    def test_show_mido_portbinding(self):
        keys = {portbindings.PROFILE: {'interface_name': 'if_name'},
                portbindings.HOST_ID: 'host'}
        with self.port_with_binding_profile() as port:
            self.assertDictSupersetOf(keys, port['port'])
            req = self.new_show_request('ports', port['port']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertDictSupersetOf(keys, res['port'])

    def test_create_mido_portbinding_no_profile_specified(self):
        with self.port() as port:
            # NOTE(yamamoto): midonet_v2 returns None where ML2 returns {}.
            self.assertEqual({}, port['port'][portbindings.PROFILE])

    def test_create_mido_portbinding_no_host_binding(self):
        # Create a binding when there is no host binding.
        # ML2 allows it and makes the port UNBOUND.
        # (Unlike midonet_v2, where it fails with 400.)
        with self.network() as net:
            args = {'port': {'tenant_id': net['network']['tenant_id'],
                             'network_id': net['network']['id'],
                             portbindings.PROFILE:
                                 {'interface_name': 'if_name'},
                             portbindings.HOST_ID: None}}
            req = self.new_create_request('ports', args, self.fmt)
            res = req.get_response(self.api)
            self.assertEqual(201, res.status_int)
            result = self.deserialize(self.fmt, res)
            self.assertDictSupersetOf({
                portbindings.PROFILE: {'interface_name': 'if_name'},
                portbindings.VIF_TYPE: portbindings.VIF_TYPE_UNBOUND,
            }, result['port'])

    def test_create_mido_portbinding_no_interface(self):
        # Create binding with no interface name.
        # ML2 allows it.
        # (Unlike midonet_v2, where a non empty profile should always have
        # a valid interface_name and otherwise fails with 400.)
        with self.network() as net:
            args = {'port': {'tenant_id': net['network']['tenant_id'],
                             'network_id': net['network']['id'],
                             portbindings.PROFILE: {'foo': ''},
                             portbindings.HOST_ID: 'host'}}
            req = self.new_create_request('ports', args, self.fmt)
            res = req.get_response(self.api)
            self.assertEqual(201, res.status_int)
            result = self.deserialize(self.fmt, res)
            self.assertDictSupersetOf({
                portbindings.PROFILE: {'foo': ''},
                portbindings.VIF_TYPE: m_const.VIF_TYPE_MIDONET,
            }, result['port'])

    def test_create_mido_portbinding_bad_interface(self):
        # Create binding with a bad interface name.  Should return an error.
        # It succeeds because our mech driver doesn't validate it.
        # (Unlike midonet_v2, where it fails with 400.)
        with self.network() as net:
            args = {'port': {'tenant_id': net['network']['tenant_id'],
                             'network_id': net['network']['id'],
                             portbindings.PROFILE: {'interface_name': ''},
                             portbindings.HOST_ID: 'host'}}
            req = self.new_create_request('ports', args, self.fmt)
            res = req.get_response(self.api)
            self.assertEqual(201, res.status_int)
            result = self.deserialize(self.fmt, res)
            self.assertDictSupersetOf({
                portbindings.PROFILE: {'interface_name': ''},
                portbindings.VIF_TYPE: m_const.VIF_TYPE_MIDONET,
            }, result['port'])

    def test_update_mido_portbinding(self):
        keys = {portbindings.HOST_ID: 'host2',
                portbindings.PROFILE: {'interface_name': 'if_name2'},
                'admin_state_up': False,
                'name': 'test_port2'}
        with self.port_with_binding_profile() as port:
            args = {
                'port': {portbindings.PROFILE: {'interface_name': 'if_name2'},
                         portbindings.HOST_ID: 'host2',
                         'admin_state_up': False,
                         'name': 'test_port2'}}
            req = self.new_update_request('ports', args, port['port']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertDictSupersetOf(keys, res['port'])

    def test_update_mido_portbinding_no_profile_specified(self):
        # Modify binding without specifying the profile.
        keys = {portbindings.HOST_ID: 'host2',
                portbindings.PROFILE: {'interface_name': 'if_name'},
                'admin_state_up': False,
                'name': 'test_port2'}
        with self.port_with_binding_profile() as port:
            args = {'port': {portbindings.HOST_ID: 'host2',
                             'admin_state_up': False,
                             'name': 'test_port2'}}
            req = self.new_update_request('ports', args, port['port']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertDictSupersetOf(keys, res['port'])

    def test_update_mido_portbinding_no_host_binding(self):
        # Update a binding when there is no host binding.
        # ML2 allows it and makes the port UNBOUND.
        # (Unlike midonet_v2, where it fails with 400.)
        with self.port() as port:
            args = {
                'port': {portbindings.PROFILE: {'interface_name': 'if_name2'}}}
            req = self.new_update_request('ports', args, port['port']['id'])
            res = req.get_response(self.api)
            self.assertEqual(200, res.status_int)
            result = self.deserialize(self.fmt, res)
            self.assertDictSupersetOf({
                portbindings.PROFILE: {'interface_name': 'if_name2'},
                portbindings.VIF_TYPE: portbindings.VIF_TYPE_UNBOUND,
            }, result['port'])

    def test_update_mido_portbinding_unbind(self):
        # Unbinding a bound port
        with self.port_with_binding_profile() as port:
            args = {'port': {portbindings.PROFILE: None}}
            req = self.new_update_request('ports', args, port['port']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            # NOTE(yamamoto): midonet_v2 returns None where ML2 returns {}.
            self.assertEqual({}, res['port'][portbindings.PROFILE])

    def test_update_mido_portbinding_unbind_already_unbound(self):
        # Unbinding an unbound port results in no-op
        with self.port() as port:
            args = {'port': {portbindings.PROFILE: None}}
            req = self.new_update_request('ports', args, port['port']['id'])
            # Success with profile set to None
            res = self.deserialize(self.fmt, req.get_response(self.api))
            # NOTE(yamamoto): midonet_v2 returns None where ML2 returns {}.
            self.assertEqual({}, res['port'][portbindings.PROFILE])

    def test_update_mido_portbinding_no_interface(self):
        # Update binding with no interface name.
        # ML2 allows it.
        # (Unlike midonet_v2, where a non empty profile should always have
        # a valid interface_name and otherwise fails with 400.)
        with self.port_with_binding_profile() as port:
            args = {
                'port': {portbindings.PROFILE: {'foo': ''}}}
            req = self.new_update_request('ports', args, port['port']['id'])
            res = req.get_response(self.api)
            self.assertEqual(200, res.status_int)
            result = self.deserialize(self.fmt, res)
            self.assertDictSupersetOf({
                portbindings.PROFILE: {'foo': ''},
                portbindings.VIF_TYPE: m_const.VIF_TYPE_MIDONET,
            }, result['port'])

    def test_update_mido_portbinding_bad_interface(self):
        # Update binding with a bad interface name.
        # It succeeds because our mech driver doesn't validate it.
        # (Unlike midonet_v2, where it fails with 400.)
        with self.port_with_binding_profile() as port:
            args = {
                'port': {portbindings.PROFILE: {'interface_name': ''}}}
            req = self.new_update_request('ports', args, port['port']['id'])
            res = req.get_response(self.api)
            self.assertEqual(200, res.status_int)
            result = self.deserialize(self.fmt, res)
            self.assertDictSupersetOf({
                portbindings.PROFILE: {'interface_name': ''},
                portbindings.VIF_TYPE: m_const.VIF_TYPE_MIDONET,
            }, result['port'])


class TestMidonetAllowedAddressPair(test_addr.TestAllowedAddressPairs,
                                    MidonetPluginML2TestCase):
    pass


class TestMidonetPortSecurity(test_psec.TestPortSecurity,
                              MidonetPluginML2TestCase):
    pass


class TestMidonetExtGwMode(test_gw_mode.ExtGwModeIntTestCase,
                           MidonetPluginML2TestCase):

    pass


class TestMidonetExtraDHCPOpts(test_dhcpopts.TestExtraDhcpOpt,
                               MidonetPluginML2TestCase):

    pass



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from midonet.neutron._i18n import _


mido_opts = [
    cfg.StrOpt('midonet_uri', default='http://localhost:8080/midonet-api',
               help=_('MidoNet API server URI. '
                      'Note that, for historical reasons, the port number '
                      'in the default value (8080) does not match the '
                      'default of the MidoNet API in MidoNet 5.0 and later, '
                      'which is 8181. '
                      'Even if you configured the MidoNet API to use port '
                      '8080, we recommend to configure this option '
                      'explicitly because the default value may change '
                      'in the future release of networking-midonet.')),
    cfg.StrOpt('username', default='admin',
               help=_('MidoNet admin username.')),
    cfg.StrOpt('password', default='passw0rd',
               secret=True,
               help=_('MidoNet admin password.')),
    cfg.StrOpt('project_id',
               default='77777777-7777-7777-7777-777777777777',
               help=_('ID of the project that MidoNet admin user '
                      'belongs to.')),
    cfg.StrOpt('tunnel_protocol', default='vxlan',
               help=_('Tunnel protocol used by Midonet. '
                      'Currently unused.')),
    cfg.StrOpt('cluster_ip', default='localhost',
               help=_('IP that the cluster service can be reached on. '
                      'Currently unused.')),
    cfg.StrOpt('cluster_port', default='8088',
               help=_('Port that the cluster service can be reached on. '
                      'Currently unused.')),
    cfg.StrOpt('client', default='midonet.neutron.client.api.MidonetApiClient',
               help=_('MidoNet client used to access MidoNet data storage. '
                      'Do not change unless you want to try the experimental '
                      'Task-based API.')),
]

cfg.CONF.register_opts(mido_opts, "MIDONET")


def list_opts():
    return [
        ('midonet', mido_opts),
    ]



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

from oslo_utils import uuidutils
import webob.exc

from neutron.tests.unit.api import test_extensions as test_ex
from neutron.tests.unit.extensions import test_l3

from midonet.neutron import extensions as midoextensions
from midonet.neutron.extensions import gateway_device
from midonet.neutron.tests.unit import test_midonet_plugin_ml2 as test_mn_ml2

FAKE_MANAGEMENT_IP = '10.0.0.3'
FAKE_MANAGEMENT_PORT = 5672
TYPE_HW_VTEP = 'hw_vtep'
TYPE_ROUTER_VTEP = 'router_vtep'
TYPE_NETWORK_VLAN = 'network_vlan'
OVSDB = 'ovsdb'
FAKE_MAC_ADDRESS = 'aa:aa:aa:aa:aa:aa'
FAKE_MAC_ADDRESS2 = 'bb:bb:bb:bb:bb:bb'
FAKE_VTEP_ADDRESS = '10.1.0.3'
FAKE_VTEP_ADDRESS2 = '10.1.0.4'
FAKE_SEG_ID = 1000
FAKE_TUNNEL_IP = '10.2.0.3'
FAKE_TUNNEL_IP2 = '10.2.0.4'
FAKE_TENANT_ID = uuidutils.generate_uuid()

DB_GATEWAY_DEVICE_PLUGIN_KLASS = 'midonet_gwdevice'
extensions_path = ':'.join(midoextensions.__path__)


class GatewayDeviceTestExtensionManager(test_l3.L3TestExtensionManager):

    def get_resources(self):
        res = super(GatewayDeviceTestExtensionManager, self).get_resources()
        return res + gateway_device.Gateway_device.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class GatewayDeviceTestCaseMixin(object):

    @contextlib.contextmanager
    def gateway_device_type_hw_vtep(self, name=TYPE_HW_VTEP,
                                    type=TYPE_HW_VTEP,
                                    management_ip=FAKE_MANAGEMENT_IP,
                                    management_port=FAKE_MANAGEMENT_PORT,
                                    management_protocol=OVSDB,
                                    tunnel_ips=None):
        if tunnel_ips is None:
            tunnel_ips = [FAKE_TUNNEL_IP]
        gw_dev = self._make_gateway_device_hw_vtep(name, type, management_ip,
                                                   management_port,
                                                   management_protocol,
                                                   tunnel_ips)
        yield gw_dev

    @contextlib.contextmanager
    def gateway_device_type_router_vtep(self, name=TYPE_ROUTER_VTEP,
                                        type=TYPE_ROUTER_VTEP,
                                        resource_id="", tunnel_ips=None):
        if tunnel_ips is None:
            tunnel_ips = [FAKE_TUNNEL_IP]
        gw_dev = self._make_gateway_device_router_vtep(name, type,
                                                       resource_id,
                                                       tunnel_ips)
        yield gw_dev

    @contextlib.contextmanager
    def gateway_device_type_network_vlan(self, name=TYPE_NETWORK_VLAN,
                                         type=TYPE_NETWORK_VLAN,
                                         resource_id=""):
        gw_dev = self._make_gateway_device_network_vlan(name, type,
                                                        resource_id)
        yield gw_dev

    def _make_gateway_device_hw_vtep(self, name, type, management_ip,
                                     management_port,
                                     management_protocol, tunnel_ips):
        res = self._create_gateway_device_hw_vtep(name, type, management_ip,
                                                  management_port,
                                                  management_protocol,
                                                  tunnel_ips)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _make_gateway_device_router_vtep(self, name, type,
                                         resource_id, tunnel_ips):
        res = self._create_gateway_device_router_vtep(name, type,
                                                      resource_id,
                                                      tunnel_ips)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _make_gateway_device_network_vlan(self, name, type,
                                          resource_id):
        res = self._create_gateway_device_network_vlan(name, type,
                                                       resource_id)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _create_gateway_device_router_vtep(self, name=TYPE_ROUTER_VTEP,
                                           type=TYPE_ROUTER_VTEP,
                                           resource_id="",
                                           tunnel_ips=None):
        data = {'gateway_device': {'name': name,
                                   'tenant_id': FAKE_TENANT_ID,
                                   'type': type,
                                   'resource_id': resource_id,
                                   'tunnel_ips': tunnel_ips or None}}
        gw_dev_req = self.new_create_request('gw/gateway_devices',
                                             data, self.fmt)
        return gw_dev_req.get_response(self.ext_api)

    def _create_gateway_device_hw_vtep(self, name=TYPE_HW_VTEP,
                                       type=TYPE_HW_VTEP,
                                       management_ip=None,
                                       management_port=None,
                                       management_protocol=None,
                                       tunnel_ips=None):
        data = {'gateway_device': {'name': name,
                                   'tenant_id': FAKE_TENANT_ID,
                                   'type': type,
                                   'management_ip': management_ip,
                                   'management_port': management_port,
                                   'management_protocol': management_protocol,
                                   'tunnel_ips': tunnel_ips or []}}
        gw_dev_req = self.new_create_request('gw/gateway_devices',
                                             data, self.fmt)
        return gw_dev_req.get_response(self.ext_api)

    def _create_gateway_device_network_vlan(self, name=TYPE_NETWORK_VLAN,
                                            type=TYPE_NETWORK_VLAN,
                                            resource_id=""):
        data = {'gateway_device': {'name': name,
                                   'tenant_id': FAKE_TENANT_ID,
                                   'type': type,
                                   'resource_id': resource_id}}
        gw_dev_req = self.new_create_request('gw/gateway_devices',
                                             data, self.fmt)
        return gw_dev_req.get_response(self.ext_api)


class GatewayDeviceTestCase(test_l3.L3NatTestCaseMixin,
                            GatewayDeviceTestCaseMixin):

    def setUp(self, plugin=None, service_plugins=None, ext_mgr=None):

        service_plugins = {
            'gateway_device_plugin_name': DB_GATEWAY_DEVICE_PLUGIN_KLASS}

        gw_dev_mgr = GatewayDeviceTestExtensionManager()
        super(GatewayDeviceTestCase, self).setUp(
            service_plugins=service_plugins, ext_mgr=gw_dev_mgr)
        self.ext_api = test_ex.setup_extensions_middleware(gw_dev_mgr)

        network = self._make_network(self.fmt, 'net1', True)
        self._subnet = self._make_subnet(self.fmt, network, "10.0.0.1",
                                         '10.0.0.0/24')
        self._subnet_id = self._subnet['subnet']['id']
        router1 = self._make_router(self.fmt, uuidutils.generate_uuid(),
                                    'router1', True)
        self._router_id = router1['router']['id']

        router2 = self._make_router(self.fmt, uuidutils.generate_uuid(),
                                    'router2', True)
        self._router_id_in_use = router2['router']['id']
        self._router_interface_action('add', self._router_id_in_use,
                                      self._subnet_id, None)

        # for network_vlan gateway device setting
        res = self._create_network(self.fmt, 'gateway_network_vlan',
                                   True)
        self._network_id = self.deserialize(self.fmt, res)['network']['id']

    def _create_remote_mac_entry(self, gw_dev_id,
                                 mac_address=FAKE_MAC_ADDRESS,
                                 vtep_address=FAKE_VTEP_ADDRESS,
                                 segmentation_id=FAKE_SEG_ID):
        data = {'remote_mac_entry': {'mac_address': mac_address,
                                     'vtep_address': vtep_address,
                                     'segmentation_id': segmentation_id,
                                     'tenant_id': FAKE_TENANT_ID}}
        gw_dev_mac_req = self.new_create_request('gw/gateway_devices/'
                                                 + gw_dev_id
                                                 + '/remote_mac_entries',
                                                 data,
                                                 self.fmt)

        return gw_dev_mac_req.get_response(self.ext_api)

    def test_create_remote_mac(self):
        expected = {'mac_address': FAKE_MAC_ADDRESS,
                    'vtep_address': FAKE_VTEP_ADDRESS,
                    'segmentation_id': FAKE_SEG_ID}
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']) as rme:
                self.assertDictSupersetOf(expected, rme['remote_mac_entry'])

    def test_create_remote_mac_on_network_vlan(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._network_id) as gw_dev:
            res = self._create_remote_mac_entry(gw_dev['gateway_device']['id'])
            self.deserialize(self.fmt, res)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_create_remote_mac_with_same_vtep_address(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']):
                with self.remote_mac_entry(gw_dev['gateway_device']['id'],
                                           mac_address=FAKE_MAC_ADDRESS2):
                    req = self.new_list_request(
                        'gw/gateway_devices/'
                        + gw_dev['gateway_device']['id']
                        + '/remote_mac_entries')
                    res = self.deserialize(
                        self.fmt, req.get_response(self.ext_api))
                    self.assertEqual(len(res['remote_mac_entries']), 2)

    def test_create_remote_mac_with_duplicate_mac_address(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']):
                res = self._create_remote_mac_entry(
                    gw_dev['gateway_device']['id'],
                    vtep_address=FAKE_VTEP_ADDRESS2)
                self.deserialize(self.fmt, res)
                self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_show_remote_mac(self):
        expected = {'mac_address': FAKE_MAC_ADDRESS,
                    'vtep_address': FAKE_VTEP_ADDRESS,
                    'segmentation_id': FAKE_SEG_ID}
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']) as rme:
                req = self.new_show_request('gw/gateway_devices/'
                                            + gw_dev['gateway_device']['id']
                                            + '/remote_mac_entries',
                                            rme['remote_mac_entry']['id'])
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                self.assertDictSupersetOf(expected, res['remote_mac_entry'])

    def test_list_remote_mac(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']):
                req = self.new_list_request('gw/gateway_devices/'
                                            + gw_dev['gateway_device']['id']
                                            + '/remote_mac_entries')
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api))
                self.assertEqual(1, len(res['remote_mac_entries']))

    def test_list_remote_mac_with_two_gateways(self):
        with self.gateway_device_type_router_vtep(
            resource_id=self._router_id) as gw_dev, \
                self.gateway_device_type_router_vtep(
                    resource_id=self._router_id_in_use,
                    tunnel_ips=[FAKE_TUNNEL_IP2]) as gw_dev2:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']) as rme:
                req = self.new_list_request('gw/gateway_devices/'
                                            + gw_dev['gateway_device']['id']
                                            + '/remote_mac_entries')
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api))
                self.assertEqual(1, len(res['remote_mac_entries']))
                self.assertEqual(rme['remote_mac_entry']['id'],
                                 res['remote_mac_entries'][0]['id'])
                req = self.new_list_request('gw/gateway_devices/'
                                            + gw_dev2['gateway_device']['id']
                                            + '/remote_mac_entries')
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api))
                self.assertEqual([], res['remote_mac_entries'])

    def test_delete_remote_mac(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']) as rme:
                req = self.new_delete_request('gw/gateway_devices/'
                                              + gw_dev['gateway_device']['id']
                                              + '/remote_mac_entries',
                                              rme['remote_mac_entry']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_remote_mac_with_wrong_gateway(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev, \
                self.gateway_device_type_router_vtep(
                    resource_id=self._router_id_in_use,
                    tunnel_ips=[FAKE_TUNNEL_IP2]) as gw_dev2:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']) as rme:
                req = self.new_delete_request('gw/gateway_devices/'
                                              + gw_dev2['gateway_device']['id']
                                              + '/remote_mac_entries',
                                              rme['remote_mac_entry']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_delete_gateway_device_with_remote_mac(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']):
                req = self.new_delete_request('gw/gateway_devices',
                                              gw_dev['gateway_device']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_create_gateway_device_with_tunnel_ips(self):
        expected = {'name': TYPE_ROUTER_VTEP,
                    'type': TYPE_ROUTER_VTEP,
                    'resource_id': self._router_id,
                    'tunnel_ips': [FAKE_TUNNEL_IP],
                    'tenant_id': FAKE_TENANT_ID}
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id,
                tunnel_ips=[FAKE_TUNNEL_IP]) as gw_dev:
            self.assertDictSupersetOf(expected, gw_dev['gateway_device'])

    def test_create_gateway_device_error_delete_neutron_resource(self):
        self.client_mock.create_gateway_device_postcommit.side_effect = (
            Exception("Fake Error"))
        self._create_gateway_device_router_vtep()
        req = self.new_list_request('gw/gateway_devices')
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertFalse(res['gateway_devices'])

    def test_create_gateway_device_error_delete_neutron_network(self):
        self.client_mock.create_gateway_device_postcommit.side_effect = (
            Exception("Fake Error"))
        self._create_gateway_device_network_vlan()
        req = self.new_list_request('gw/gateway_devices')
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertFalse(res['gateway_devices'])

    def test_update_gateway_device_error_rollback_neutron_resource(self):
        self.client_mock.update_gateway_device_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.gateway_device_type_router_vtep(
            resource_id=self._router_id,
                tunnel_ips=[FAKE_TUNNEL_IP]) as gw_dev:
            data = {'gateway_device': {'tunnel_ips': [FAKE_TUNNEL_IP2]}}
            req = self.new_update_request('gw/gateway_devices',
                                          data,
                                          gw_dev['gateway_device']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            # ensure TunnelIPs are not changed.
            expected = {'tunnel_ips': [FAKE_TUNNEL_IP]}
            req = self.new_show_request('gw/gateway_devices',
                                        gw_dev['gateway_device']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['gateway_device'])

    def test_create_remote_mac_entry_error_delete_neutron_resource(self):
        (self.client_mock.create_gateway_device_remote_mac_entry_postcommit.
         side_effect) = Exception("Fake Error")
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            try:
                with self.remote_mac_entry(gw_dev['gateway_device']['id']):
                    self.assertTrue(False)
            except webob.exc.HTTPClientError:
                pass
            req = self.new_list_request('gw/gateway_devices/'
                                        + gw_dev['gateway_device']['id']
                                        + '/remote_mac_entries')
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertFalse(res['remote_mac_entries'])

    def test_delete_gateway_device_error_delete_neutron_resource(self):
        self.client_mock.delete_gateway_device_postcommit.side_effect = (
            Exception("Fake Error"))
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            req = self.new_delete_request('gw/gateway_devices',
                                          gw_dev['gateway_device']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            # check the resource deleted in Neutron DB
            req = self.new_list_request('gw/gateway_devices')
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            self.assertFalse(res['gateway_devices'])

    def test_delete_remote_mac_entry_error_delete_neutron_resource(self):
        (self.client_mock.delete_gateway_device_remote_mac_entry_postcommit.
         side_effect) = Exception("Fake Error")
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.remote_mac_entry(gw_dev['gateway_device']['id']) as rme:
                req = self.new_delete_request('gw/gateway_devices/'
                                              + gw_dev['gateway_device']['id']
                                              + '/remote_mac_entries',
                                              rme['remote_mac_entry']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPInternalServerError.code,
                                 res.status_int)
                # check the resource deleted in Neutron DB
                req = self.new_list_request('gw/gateway_devices')
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                self.assertEqual(
                    [],
                    res['gateway_devices'][0]['remote_mac_entries'])

    def _make_remote_mac_entry(self, gw_dev_id, mac_address=FAKE_MAC_ADDRESS,
                               vtep_address=FAKE_VTEP_ADDRESS,
                               segmentation_id=FAKE_SEG_ID):
        res = self._create_remote_mac_entry(gw_dev_id, mac_address,
                                            vtep_address,
                                            segmentation_id)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    @contextlib.contextmanager
    def remote_mac_entry(self, gw_dev_id, mac_address=FAKE_MAC_ADDRESS,
                         vtep_address=FAKE_VTEP_ADDRESS,
                         segmentation_id=FAKE_SEG_ID):
        rme = self._make_remote_mac_entry(gw_dev_id, mac_address, vtep_address,
                                          segmentation_id)
        yield rme

    def test_create_gateway_device_hw_vtep(self):
        expected = {'name': TYPE_HW_VTEP,
                    'type': TYPE_HW_VTEP,
                    'management_ip': FAKE_MANAGEMENT_IP,
                    'management_port': FAKE_MANAGEMENT_PORT,
                    'management_protocol': OVSDB,
                    'tenant_id': FAKE_TENANT_ID}
        with self.gateway_device_type_hw_vtep() as gw_dev:
            self.assertDictSupersetOf(expected, gw_dev['gateway_device'])

    def test_create_gateway_device_router_vtep(self):
        expected = {'name': TYPE_ROUTER_VTEP,
                    'type': TYPE_ROUTER_VTEP,
                    'resource_id': self._router_id,
                    'tenant_id': FAKE_TENANT_ID}
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            self.assertDictSupersetOf(expected, gw_dev['gateway_device'])

    def test_create_gateway_device_network_vlan(self):
        expected = {'name': TYPE_NETWORK_VLAN,
                    'type': TYPE_NETWORK_VLAN,
                    'resource_id': self._network_id,
                    'tenant_id': FAKE_TENANT_ID}
        with self.gateway_device_type_network_vlan(
                resource_id=self._network_id) as gw_dev:
            self.assertDictSupersetOf(expected, gw_dev['gateway_device'])

    def test_create_gateway_device_router_vtep_not_found(self):
        res = self._create_gateway_device_router_vtep(resource_id='a')
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_create_gateway_device_network_vlan_not_found(self):
        res = self._create_gateway_device_network_vlan(resource_id='a')
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_create_gateway_device_with_multiple_tunnel_ips(self):
        res = self._create_gateway_device_router_vtep(
            resource_id=self._router_id,
            tunnel_ips=[FAKE_TUNNEL_IP, FAKE_TUNNEL_IP2])
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_gateway_device_without_tunnel_ips(self):
        res = self._create_gateway_device_router_vtep(
            resource_id=self._router_id,
            tunnel_ips=[])
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_update_gateway_device_with_multiple_tunnel_ips(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            data = {
                'gateway_device': {
                    'tunnel_ips': [
                        FAKE_TUNNEL_IP,
                        FAKE_TUNNEL_IP2
                    ]
                }
            }
            gw_dev_req = self.new_update_request(
                'gw/gateway_devices',
                data,
                gw_dev['gateway_device']['id'])
            res = gw_dev_req.get_response(self.ext_api)
            self.deserialize(self.fmt, res)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_update_gateway_device_without_tunnel_ips(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            data = {'gateway_device': {'tunnel_ips': []}}
            gw_dev_req = self.new_update_request(
                'gw/gateway_devices',
                data,
                gw_dev['gateway_device']['id'])
            res = gw_dev_req.get_response(self.ext_api)
            self.deserialize(self.fmt, res)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_gateway_device_router_with_duplicate_router(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            res = self._create_gateway_device_router_vtep(
                resource_id=gw_dev['gateway_device']['resource_id'])
            self.deserialize(self.fmt, res)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_create_gateway_device_network_with_duplicate_network(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._network_id) as gw_dev:
            res = self._create_gateway_device_network_vlan(
                resource_id=gw_dev['gateway_device']['resource_id'])
            self.deserialize(self.fmt, res)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_create_gateway_device_hw_vtep_without_management_ip(self):
        res = self._create_gateway_device_hw_vtep(
            management_port=FAKE_MANAGEMENT_PORT,
            management_protocol=OVSDB,
            tunnel_ips=[FAKE_TUNNEL_IP])
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_gateway_device_hw_vtep_without_management_port(self):
        res = self._create_gateway_device_hw_vtep(
            management_ip=FAKE_MANAGEMENT_IP,
            management_protocol=OVSDB,
            tunnel_ips=[FAKE_TUNNEL_IP])
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_gateway_device_hw_vtep_without_management_protocol(self):
        res = self._create_gateway_device_hw_vtep(
            management_ip=FAKE_MANAGEMENT_IP,
            management_port=FAKE_MANAGEMENT_PORT,
            tunnel_ips=[FAKE_TUNNEL_IP])
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)

    def test_create_gateway_device_hw_vtep_without_tunnel_ips(self):
        res = self._create_gateway_device_hw_vtep(
            management_ip=FAKE_MANAGEMENT_IP,
            management_port=FAKE_MANAGEMENT_PORT,
            management_protocol=OVSDB)
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_gateway_device_hw_vtep_with_string_management_port(self):
        res = self._create_gateway_device_hw_vtep(
            management_ip=FAKE_MANAGEMENT_IP,
            management_port=str(FAKE_MANAGEMENT_PORT),
            management_protocol=OVSDB,
            tunnel_ips=[FAKE_TUNNEL_IP])
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)

    def test_create_gateway_device_router_vtep_without_resource_id(self):
        res = self._create_gateway_device_router_vtep()
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_gateway_device_vlan_network_without_resource_id(self):
        res = self._create_gateway_device_network_vlan()
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_delete_gateway_device(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            req = self.new_delete_request('gw/gateway_devices',
                                          gw_dev['gateway_device']['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_router_in_use(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id):
            req = self.new_delete_request('routers',
                                          self._router_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)

    def test_delete_network_in_use(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._network_id):
            req = self.new_delete_request('networks',
                                          self._network_id)
            res = req.get_response(self.api)
            self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)
            req = self.new_show_request('networks', self._network_id)
            res = req.get_response(self.api)
            self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

    def test_update_gateway_device(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            data = {'gateway_device': {'name': 'new_name'}}
            req = self.new_update_request('gw/gateway_devices',
                                          data,
                                          gw_dev['gateway_device']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertEqual(data['gateway_device']['name'],
                             res['gateway_device']['name'])

    def test_update_gateway_device_tunnel_ips(self):
        with self.gateway_device_type_router_vtep(
            resource_id=self._router_id,
                tunnel_ips=[FAKE_TUNNEL_IP]) as gw_dev:
            data = {'gateway_device': {'tunnel_ips': [FAKE_TUNNEL_IP2]}}
            req = self.new_update_request('gw/gateway_devices',
                                          data,
                                          gw_dev['gateway_device']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertEqual(data['gateway_device']['tunnel_ips'],
                             res['gateway_device']['tunnel_ips'])

    def test_show_gateway_device_hw_vtep(self):
        expected = {'name': TYPE_HW_VTEP,
                    'type': TYPE_HW_VTEP,
                    'management_ip': FAKE_MANAGEMENT_IP,
                    'management_port': FAKE_MANAGEMENT_PORT,
                    'management_protocol': OVSDB,
                    'tenant_id': FAKE_TENANT_ID}
        with self.gateway_device_type_hw_vtep() as gw_dev:
            req = self.new_show_request('gw/gateway_devices',
                                        gw_dev['gateway_device']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['gateway_device'])

    def test_show_gateway_device_router_vtep(self):
        expected = {'name': TYPE_ROUTER_VTEP,
                    'type': TYPE_ROUTER_VTEP,
                    'resource_id': self._router_id,
                    'tenant_id': FAKE_TENANT_ID}
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            req = self.new_show_request('gw/gateway_devices',
                                        gw_dev['gateway_device']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['gateway_device'])

    def test_show_gateway_device_network_vlan(self):
        expected = {'name': TYPE_NETWORK_VLAN,
                    'type': TYPE_NETWORK_VLAN,
                    'resource_id': self._network_id,
                    'tenant_id': FAKE_TENANT_ID}
        with self.gateway_device_type_network_vlan(
                resource_id=self._network_id) as gw_dev:
            req = self.new_show_request('gw/gateway_devices',
                                        gw_dev['gateway_device']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertDictSupersetOf(expected, res['gateway_device'])

    def test_list_gateway_devices(self):
        with self.gateway_device_type_router_vtep(resource_id=self._router_id):
            with self.gateway_device_type_hw_vtep(
                    tunnel_ips=[FAKE_TUNNEL_IP2]):
                req = self.new_list_request('gw/gateway_devices')
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api))
                self.assertEqual(2, len(res['gateway_devices']))


class GatewayDeviceTestCaseWithML2(GatewayDeviceTestCase,
                                   test_mn_ml2.MidonetPluginML2TestCase):
    pass



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""rename tenant to project

Revision ID: 27e6e3451f22
Revises: 7234c9979bf2
Create Date: 2016-07-15 01:26:56.968744

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine import reflection

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = '27e6e3451f22'
down_revision = '7234c9979bf2'
depends_on = ('9f3992811410',)

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.NEWTON]


_INSPECTOR = None


def get_inspector():
    """Reuse inspector"""

    global _INSPECTOR

    if _INSPECTOR:
        return _INSPECTOR

    else:
        bind = op.get_bind()
        _INSPECTOR = reflection.Inspector.from_engine(bind)

    return _INSPECTOR


def get_tables():
    """Returns hardcoded list of tables which have ``tenant_id`` column.

    The list is hard-coded to match the state of the schema when this upgrade
    script is run.
    """

    tables = [
        'midonet_tasks',
        'midonet_gateway_devices',
        'midonet_logging_resources',
        'midonet_firewall_logs',
    ]

    return tables


def get_columns(table):
    """Returns list of columns for given table."""
    inspector = get_inspector()
    return inspector.get_columns(table)


def get_data():
    """Returns combined list of tuples: [(table, column)].

    The list is built from tables with a tenant_id column.
    """

    output = []
    tables = get_tables()
    for table in tables:
        columns = get_columns(table)

        for column in columns:
            if column['name'] == 'tenant_id':
                output.append((table, column))

    return output


def alter_column(table, column):
    old_name = 'tenant_id'
    new_name = 'project_id'

    op.alter_column(
        table_name=table,
        column_name=old_name,
        new_column_name=new_name,
        existing_type=column['type'],
        existing_nullable=column['nullable']
    )


def upgrade():
    data = get_data()
    for table, column in data:
        alter_column(table, column)


def contract_creation_exceptions():
    """Special migration for the blueprint to support Keystone V3.

    We drop all tenant_id columns and create project_id columns instead.
    """

    return {
        sa.Column: ['.'.join([table, 'project_id']) for table in get_tables()],
        sa.Index: get_tables()
    }



# Copyright (C) 2016 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import mock
from oslo_utils import uuidutils
import webob.exc

from neutron.db import servicetype_db as sdb
from neutron.services import provider_configuration as provconf
from neutron.tests.unit.api import test_extensions as test_ex
from neutron_taas.common import constants as taas_const
from neutron_taas.extensions import taas as ext_taas
from neutron_taas.tests.unit.services.taas import test_taas_plugin  # noqa

from midonet.neutron.tests.unit import test_midonet_plugin_ml2 as test_mn_ml2

# Generate uuids
TENANT1 = uuidutils.generate_uuid()
TENANT2 = uuidutils.generate_uuid()

TAAS_PREFIX = '/taas'

DB_TAAS_PLUGIN_KLASS = ('neutron_taas.services.taas.taas_plugin.TaasPlugin')
MN_TAAS_DRIVER_KLASS = ('midonet.neutron.services.taas.service_drivers.'
                        'taas_midonet.MidonetTaasDriver')


class TaasExtensionManager(object):

    def get_resources(self):
        return ext_taas.Taas.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class TestMidonetTaasCaseMixin(object):
    resource_prefix_map = dict(
        (k, TAAS_PREFIX)
        for k in ext_taas.RESOURCE_ATTRIBUTE_MAP.keys()
    )

    def setUp(self, plugin=None, service_plugins=None, ext_mgr=None):
        ext_mgr = TaasExtensionManager()
        service_plugins = {'taas_plugin_name': DB_TAAS_PLUGIN_KLASS}
        taas_provider = (taas_const.TAAS + ':Midonet:' + MN_TAAS_DRIVER_KLASS
                         + ':default')
        mock.patch.object(provconf.NeutronModule, 'service_providers',
                          return_value=[taas_provider]).start()
        manager = sdb.ServiceTypeManager.get_instance()
        manager.add_provider_configuration(
            taas_const.TAAS, provconf.ProviderConfiguration())
        super(TestMidonetTaasCaseMixin,
              self).setUp(service_plugins=service_plugins,
                          ext_mgr=ext_mgr)
        self.ext_api = test_ex.setup_extensions_middleware(ext_mgr)

    def _create_tap_service(self, port_id, name=None, tenant_id=None):
        t_s = {'tap_service': {'port_id': port_id}}
        if name:
            t_s['tap_service']['name'] = name
        if tenant_id:
            t_s['tap_service']['tenant_id'] = tenant_id

        ts_req = self.new_create_request('tap_services', t_s, self.fmt)
        return ts_req.get_response(self.ext_api)

    def _create_tap_flow(self, tap_service_id, source_port, direction,
                         name=None, tenant_id=None):
        t_f = {
            'tap_flow': {
                'tap_service_id': tap_service_id,
                'source_port': source_port,
                'direction': direction
            }
        }
        if name:
            t_f['tap_flow']['name'] = name
        if tenant_id:
            t_f['tap_flow']['tenant_id'] = tenant_id

        tf_req = self.new_create_request('tap_flows', t_f, self.fmt)
        return tf_req.get_response(self.ext_api)

    def _make_tap_service(self, port_id, name=None, tenant_id=None):
        res = self._create_tap_service(port_id, name, tenant_id)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _make_tap_flow(self, tap_service_id, source_port, direction,
                       name=None, tenant_id=None):
        res = self._create_tap_flow(tap_service_id,
                                    source_port,
                                    direction,
                                    name, tenant_id)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    @contextlib.contextmanager
    def tap_service(self, port_id, name='tap_service1', tenant_id=TENANT1):
        ts = self._make_tap_service(port_id, name, tenant_id)
        yield ts

    @contextlib.contextmanager
    def tap_flow(self, tap_service_id, source_port,
                 direction='BOTH', name='tap_flow1', tenant_id=TENANT1):
        tf = self._make_tap_flow(tap_service_id, source_port,
                                 direction, name, tenant_id)
        yield tf

    def test_create_tap_service(self):
        with self.port(tenant_id=TENANT1) as port:
            dist_port = port['port']
            with self.tap_service(port_id=dist_port['id']) as t_s:
                ts = t_s['tap_service']
                self.assertEqual(dist_port['id'], ts['port_id'])
        self.client_mock.create_tap_service.assert_called_with(mock.ANY, ts)

    def test_create_tap_service_error_delete_neutron_resource(self):
        self.client_mock.create_tap_service.side_effect = (
            Exception("Fake Error"))
        with self.port(tenant_id=TENANT1) as port:
            dist_port = port['port']
            res = self._create_tap_service(port_id=dist_port['id'],
                                           tenant_id=TENANT1)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            req = self.new_list_request('tap_services')
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertFalse(res['tap_services'])

    def test_delete_tap_service(self):
        with self.port(tenant_id=TENANT1) as port:
            dist_port = port['port']
            with self.tap_service(port_id=dist_port['id']) as t_s:
                ts = t_s['tap_service']
                req = self.new_delete_request('tap_services', ts['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code,
                                 res.status_int)
        self.client_mock.delete_tap_service.assert_called_with(mock.ANY,
                                                               ts['id'])

    def test_delete_tap_service_error_delete_neutron_resource(self):
        self.client_mock.delete_tap_service.side_effect = (
            Exception("Fake Error"))
        with self.port(tenant_id=TENANT1) as port:
            dist_port = port['port']
            with self.tap_service(port_id=dist_port['id']) as t_s:
                req = self.new_delete_request('tap_services',
                                              t_s['tap_service']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPInternalServerError.code,
                                 res.status_int)
                req = self.new_list_request('tap_services')
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                self.assertFalse(res['tap_services'])

    @contextlib.contextmanager
    def create_tap_service_and_tap_flow(self):
        with self.port(tenant_id=TENANT1) as port1, \
                self.port(tenant_id=TENANT1) as port2:
            dist_port = port1['port']
            source_port = port2['port']
            with self.tap_service(port_id=dist_port['id']) as t_s:
                ts = t_s['tap_service']
                with self.tap_flow(tap_service_id=ts['id'],
                                   source_port=source_port['id']) as t_f:
                    tf = t_f['tap_flow']
                    self.assertEqual(source_port['id'], tf['source_port'])
                    yield tf

    def test_create_tap_flow(self):
        with self.create_tap_service_and_tap_flow() as tf:
            pass
        self.client_mock.create_tap_flow.assert_called_with(mock.ANY, tf)

    def test_create_tap_flow_error_delete_neutron_resource(self):
        self.client_mock.create_tap_flow.side_effect = (
            Exception("Fake Error"))
        with self.port(tenant_id=TENANT1) as port1, \
                self.port(tenant_id=TENANT1) as port2:
            dist_port = port1['port']
            source_port = port2['port']
            with self.tap_service(port_id=dist_port['id']) as t_s:
                ts = t_s['tap_service']
                res = self._create_tap_flow(ts['id'], source_port['id'],
                                            'BOTH', 'test', tenant_id=TENANT1)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            req = self.new_list_request('tap_flows')
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertFalse(res['tap_flows'])

    def test_delete_tap_flow(self):
        with self.create_tap_service_and_tap_flow() as tf:
            req = self.new_delete_request('tap_flows', tf['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        self.client_mock.delete_tap_flow.assert_called_with(mock.ANY, tf['id'])

    def test_delete_tap_flow_error_delete_neutron_resource(self):
        self.client_mock.delete_tap_flow.side_effect = (
            Exception("Fake Error"))
        with self.create_tap_service_and_tap_flow() as tf:
            req = self.new_delete_request('tap_flows', tf['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPInternalServerError.code,
                             res.status_int)
            req = self.new_list_request('tap_flows')
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
            self.assertFalse(res['tap_flows'])

    def test_delete_tap_service_with_tap_flow(self):
        with self.create_tap_service_and_tap_flow() as tf:
            req = self.new_show_request('tap_services', tf['tap_service_id'])
            res = req.get_response(self.ext_api)
            t_s = self.deserialize(self.fmt, req.get_response(self.ext_api))
            ts = t_s['tap_service']
            req = self.new_delete_request('tap_services', ts['id'])
            res = req.get_response(self.ext_api)
            self.assertEqual(webob.exc.HTTPNoContent.code,
                             res.status_int)
        self.client_mock.delete_tap_flow.assert_called_with(mock.ANY, tf['id'])
        self.client_mock.delete_tap_service.assert_called_with(mock.ANY,
                                                               ts['id'])


class TestMidonetTaasCaseML2(TestMidonetTaasCaseMixin,
                             test_mn_ml2.MidonetPluginML2TestCase):
    pass



# Copyright 2017 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os_vif import objects
from os_vif import plugin
from oslo_concurrency import processutils

from midonet.os_vif import linux_net
from midonet.os_vif import privsep


class MidoNetPlugin(plugin.PluginBase):

    def describe(self):
        return objects.host_info.HostPluginInfo(
            plugin_name="midonet",
            vif_info=[
                objects.host_info.HostVIFInfo(
                    vif_object_name=objects.vif.VIFGeneric.__name__,
                    min_version="1.0",
                    max_version="1.0"),
            ])

    def plug(self, vif, instance_info):
        linux_net.create_tap_dev(vif.vif_name)
        _bind_port(vif.id, vif.vif_name)

    def unplug(self, vif, instance_info):
        _unbind_port(vif.id)
        linux_net.delete_net_dev(vif.vif_name)


@privsep.mm_ctl.entrypoint
def _bind_port(port_id, ifname):
    cmd = ('mm-ctl', '--bind-port', port_id, ifname)
    processutils.execute(*cmd)


@privsep.mm_ctl.entrypoint
def _unbind_port(port_id):
    cmd = ('mm-ctl', '--unbind-port', port_id)
    processutils.execute(*cmd)



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six

from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils

LOG = logging.getLogger(__name__)


def load_client(conf):
    try:
        client = importutils.import_object(conf.client, conf)
        LOG.debug("Loaded midonet client '%(client)s'", {'client': client})
        return client
    except ImportError:
        with excutils.save_and_reraise_exception():
            LOG.exception("Error loading midonet client '%(client)s'",
                          {'client': conf.client})


@six.add_metaclass(abc.ABCMeta)
class MidonetClientBase(object):
    """Neutron MidoNet client base class.

    This class abstracts the communication between Neutron and MidoNet as there
    may be multiple ways to do so.  All MidoNet clients intended to be used
    from the Neutron plugin should extend this class.
    """

    def initialize(self):
        pass

    def create_network_precommit(self, context, network):
        pass

    def create_network_postcommit(self, network):
        pass

    def update_network_precommit(self, context, network_id, network):
        pass

    def update_network_postcommit(self, network_id, network):
        pass

    def delete_network_precommit(self, context, network_id):
        pass

    def delete_network_postcommit(self, network_id):
        pass

    def create_subnet_precommit(self, context, subnet):
        pass

    def create_subnet_postcommit(self, subnet):
        pass

    def update_subnet_precommit(self, context, subnet_id, subnet):
        pass

    def update_subnet_postcommit(self, subnet_id, subnet):
        pass

    def delete_subnet_precommit(self, context, subnet_id):
        pass

    def delete_subnet_postcommit(self, subnet_id):
        pass

    def create_port_precommit(self, context, port):
        pass

    def create_port_postcommit(self, port):
        pass

    def update_port_precommit(self, context, port_id, port):
        pass

    def update_port_postcommit(self, port_id, port):
        pass

    def delete_port_precommit(self, context, port_id):
        pass

    def delete_port_postcommit(self, port_id):
        pass

    def create_router_precommit(self, context, router):
        pass

    def create_router_postcommit(self, router):
        pass

    def update_router_precommit(self, context, router_id, router):
        pass

    def update_router_postcommit(self, router_id, router):
        pass

    def delete_router_precommit(self, context, router_id):
        pass

    def delete_router_postcommit(self, router_id):
        pass

    def add_router_interface_precommit(self, context, router_id,
                                       interface_info):
        pass

    def add_router_interface_postcommit(self, router_id, interface_info):
        pass

    def remove_router_interface_precommit(self, context, router_id,
                                          interface_info):
        pass

    def remove_router_interface_postcommit(self, router_id, interface_info):
        pass

    def create_floatingip_precommit(self, context, floatingip):
        pass

    def create_floatingip_postcommit(self, floatingip):
        pass

    def update_floatingip_precommit(self, context, floatingip_id, floatingip):
        pass

    def update_floatingip_postcommit(self, floatingip_id, floatingip):
        pass

    def delete_floatingip_precommit(self, context, floatingip_id):
        pass

    def delete_floatingip_postcommit(self, floatingip_id):
        pass

    def create_security_group_precommit(self, context, security_group):
        pass

    def create_security_group_postcommit(self, security_group):
        pass

    def update_security_group_precommit(self, context, security_group_id,
                                        security_group):
        pass

    def update_security_group_postcommit(self, security_group_id,
                                         security_group):
        pass

    def delete_security_group_precommit(self, context, security_group_id):
        pass

    def delete_security_group_postcommit(self, security_group_id):
        pass

    def create_security_group_rule_precommit(self, context,
                                             security_group_rule):
        pass

    def create_security_group_rule_postcommit(self, security_group_rule):
        pass

    def create_security_group_rule_bulk_precommit(self, context,
                                                  security_group_rules):
        pass

    def create_security_group_rule_bulk_postcommit(self, security_group_rules):
        pass

    def delete_security_group_rule_precommit(self, context,
                                             security_group_rule_id):
        pass

    def delete_security_group_rule_postcommit(self, security_group_rule_id):
        pass

    # Agent membership extension

    def create_agent_membership_precommit(self, context, agent_membership):
        pass

    def create_agent_membership_postcommit(self, agent_membership):
        pass

    def delete_agent_membership_precommit(self, context, agent_membership_id):
        pass

    def delete_agent_membership_postcommit(self, agent_membership_id):
        pass

    # LBaaS methods - these are not inside a DB transaction.  DB updates as
    # well as blocking API calls are allowed.

    def create_vip(self, context, vip):
        pass

    def update_vip(self, context, vip_id, new_vip):
        pass

    def delete_vip(self, context, vip_id):
        pass

    def create_pool(self, context, pool):
        pass

    def update_pool(self, context, pool_id, pool):
        pass

    def delete_pool(self, context, pool_id):
        pass

    def create_member(self, context, member):
        pass

    def update_member(self, context, member_id, member):
        pass

    def delete_member(self, context, member_id):
        pass

    def create_health_monitor(self, context, health_monitor):
        pass

    def update_health_monitor(self, context, health_monitor_id,
                              health_monitor):
        pass

    def delete_health_monitor(self, context, health_monitor_id):
        pass

    # FWaaS methods - these are not called inside DB transaction.

    def create_firewall(self, context, firewall):
        pass

    def delete_firewall(self, context, firewall):
        pass

    def update_firewall(self, context, firewall):
        pass

    # VPNaaS methods

    def create_vpn_service(self, context, vpn_service):
        pass

    def update_vpn_service(self, context, vpn_service_id, vpn_service):
        pass

    def delete_vpn_service(self, context, vpn_service_id):
        pass

    def create_ipsec_site_conn(self, context, ipsec_site_conn):
        pass

    def update_ipsec_site_conn(
            self, context, ipsec_site_conn_id, ipsec_site_conn):
        pass

    def delete_ipsec_site_conn(self, context, ipsec_site_conn_id):
        pass

    # Gateway Device extensions

    def create_gateway_device_precommit(self, context, gw_dev):
        pass

    def create_gateway_device_postcommit(self, gw_dev):
        pass

    def update_gateway_device_precommit(self, context, gw_dev_id, gw_dev):
        pass

    def update_gateway_device_postcommit(self, gw_dev_id, gw_dev):
        pass

    def delete_gateway_device_precommit(self, context, gw_dev_id):
        pass

    def delete_gateway_device_postcommit(self, gw_dev_id):
        pass

    def create_gateway_device_remote_mac_entry_precommit(self, context,
                                                         mac_entry):
        pass

    def create_gateway_device_remote_mac_entry_postcommit(self, mac_entry):
        pass

    def delete_gateway_device_remote_mac_entry_precommit(self, context,
                                                         mac_entry_id):
        pass

    def delete_gateway_device_remote_mac_entry_postcommit(self, mac_entry_id):
        pass

    # L2 Gateway extensions

    def create_l2_gateway_connection(self, context, l2_gw_conn):
        pass

    def delete_l2_gateway_connection(self, context, l2_gw_conn_id):
        pass

    # BGP extensions

    def update_bgp_speaker_precommit(self, context,
                                     bgp_speaker_id, bgp_speaker):
        pass

    def update_bgp_speaker_postcommit(self, bgp_speaker_id, bgp_speaker):
        pass

    def create_bgp_peer_precommit(self, context, bgp_peer):
        pass

    def create_bgp_peer_postcommit(self, bgp_peer):
        pass

    def update_bgp_peer_precommit(self, context, bgp_peer_id, bgp_peer):
        pass

    def update_bgp_peer_postcommit(self, bgp_peer_id, bgp_peer):
        pass

    def delete_bgp_peer_precommit(self, context, bgp_peer_id):
        pass

    def delete_bgp_peer_postcommit(self, bgp_peer_id):
        pass

    # Logging Resource extensions

    def update_logging_resource_precommit(
            self, context, logging_resource_id, logging_resource):
        pass

    def update_logging_resource_postcommit(self, logging_resource_id,
                                           logging_resource):
        pass

    def delete_logging_resource_precommit(self, context, logging_resource_id):
        pass

    def delete_logging_resource_postcommit(self, logging_resource_id):
        pass

    def create_firewall_log_precommit(self, context, firewall_log):
        pass

    def create_firewall_log_postcommit(self, firewall_log):
        pass

    def update_firewall_log_precommit(self, context,
                                      firewall_log_id, firewall_log):
        pass

    def update_firewall_log_postcommit(self, firewall_log_id, firewall_log):
        pass

    def delete_firewall_log_precommit(self, context, firewall_log_id):
        pass

    def delete_firewall_log_postcommit(self, firewall_log_id):
        pass

    # Tap as a service

    def create_tap_flow(self, context, tap_flow):
        pass

    def delete_tap_flow(self, context, tap_flow_id):
        pass

    def create_tap_service(self, context, tap_service):
        pass

    def delete_tap_service(self, context, tap_service_id):
        pass

    # QoS
    def create_qos_policy(self, context, policy):
        pass

    def update_qos_policy(self, context, policy):
        pass

    def delete_qos_policy(self, context, policy):
        pass

    # MidoNet Pool Member
    def get_pool_member(self, context, member_id):
        pass

    # REVISIT(yamamoto): LBaaS v2



# Copyright 2016 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add logging resource

Revision ID: 9f3992811410
Revises: f8b289f2644f
Create Date: 2016-06-06 09:39:35.381427

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = '9f3992811410'
down_revision = 'f8b289f2644f'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [
    migration.MITAKA, migration.NEWTON, migration.OCATA,
    migration.PIKE
]


def upgrade():
    op.create_table('midonet_logging_resources',
                    sa.Column('id', sa.String(length=36), primary_key=True),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('description', sa.String(length=1024),
                              nullable=True),
                    sa.Column('tenant_id', sa.String(length=255),
                              nullable=True),
                    sa.Column('enabled', sa.Boolean(), nullable=False))

    op.create_table('midonet_firewall_logs',
                    sa.Column('id', sa.String(length=36), primary_key=True),
                    sa.Column('logging_resource_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('tenant_id', sa.String(length=255),
                              nullable=True),
                    sa.Column('description', sa.String(length=1024),
                              nullable=True),
                    sa.Column('fw_event', sa.String(length=255),
                              nullable=False),
                    sa.Column('firewall_id', sa.String(length=36),
                              nullable=False),
                    sa.ForeignKeyConstraint(['logging_resource_id'],
                                            ['midonet_logging_resources.id'],
                                            ondelete='CASCADE'),
                    sa.ForeignKeyConstraint(['firewall_id'],
                                            ['firewalls.id'],
                                            ondelete='CASCADE'))



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import router_interface_fip
from neutron_lib.api import extensions
from neutron_lib import exceptions as n_exc

from midonet.neutron._i18n import _


class RouterInterfaceInUseAsGatewayByFloatingIP(n_exc.InUse):
    message = _("Router interface for subnet %(subnet_id)s on router "
                "%(router_id)s cannot be deleted, as it is required "
                "by one or more floating IPs as a gateway.")


class Routerinterfacefip(extensions.APIExtensionDescriptor):
    """Router interface FIP extension."""

    api_definition = router_interface_fip



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# MidoNet Network Type constants
TYPE_UPLINK = 'uplink'
TYPE_MIDONET = 'midonet'

MIDONET_NET_TYPES = [TYPE_MIDONET, TYPE_UPLINK]

# Midonet VIF TYPE
VIF_TYPE_MIDONET = 'midonet'

# Neutron well-known service type constants:
GATEWAY_DEVICE = "GATEWAY_DEVICE"
LOGGING_RESOURCE = "LOGGING_RESOURCE"

# for Midonet L2 Gateway
MAX_VXLAN_VNI = 16777215
MIDONET_L2GW_PROVIDER = "midonet"

# (Kengo) We define constants until
# upstream will deal with router status.
ROUTER_STATUS_ACTIVE = "ACTIVE"
ROUTER_STATUS_ERROR = "ERROR"

# for bgp dynamic router extension
LOGICAL_ROUTER = 'logical_router'



#
# Copyright 2016 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utility functions used for REST APIs"""


from __future__ import print_function

import ConfigParser
import os
from oslo_serialization import jsonutils
import requests

from rally.common import logging
from rally.task import scenario

LOG = logging.getLogger(__name__)


class RequestScenario(scenario.Scenario):
    """Base class for Request scenarios"""

    @staticmethod
    def get_credentials():
        """Retrieves credentials to connect to neutron from 'midonetrc' file"""
        try:
            parser = ConfigParser.SafeConfigParser()
            parser.read(os.path.join(os.path.expanduser('~stack'),
                                     '.midonetrc'))
            return parser
        except Exception:
            raise

    @staticmethod
    def _get_token():
        """Generates authentication token

        This method executes login API of MidoNet and generates authentication
        token. This token is generated by executing /login API providing
        base64 value of credentials
        """

        import base64
        auth_token = None
        creds = RequestScenario.get_credentials()
        credentials = base64.standard_b64encode(creds.get('cli', 'username') +
                                                ":" +
                                                creds.get('cli', 'password'))

        header = {'X-Auth-Project': creds.get('cli', 'project_id'),
                  'Authorization': 'Basic %s' % credentials}
        # Execute login API
        response = requests.request("POST", creds.get('cli', 'api_url') +
                                    "/login", headers=header)
        if response.status_code == requests.codes.ok:
            auth_token = response.json()['key']
        else:
            LOG.error("Error for the operation is: %s" % response.text)
            raise ValueError("Expected HTTP request code is 200 actual %s"
                             % response.status_code)
        return auth_token

    def request_get_all(self, api, header):
        """Executes GET request to get all values for a resource"""

        creds = RequestScenario.get_credentials()
        url = creds.get('cli', 'api_url') + '/' + api
        response = requests.request("GET", url, headers=header)
        if response.status_code == requests.codes.ok:
            LOG.debug("Resource retrieval is successful for: '%s'" % api)
            return response.json()
        elif response.status_code == 401:
            return response.status_code
        else:
            LOG.error("Error for the operation is: %s" % response.text)
            raise ValueError("Expected HTTP request code is 200 actual %s"
                             % response.status_code)

    def request_get_resource(self, api, header):
        """Executes GET request to get information of a particular resource"""

        creds = RequestScenario.get_credentials()
        url = creds.get('cli', 'api_url') + '/' + api
        response = requests.request("GET", url, headers=header)
        if response.status_code == requests.codes.ok:
            LOG.debug("Resource retrieval by ID is successful for: '%s'" % api)
            return response.status_code
        elif response.status_code == 401:
            return response.status_code
        else:
            LOG.error("Error for the operation is: %s" % response.text)
            raise ValueError("Expected HTTP request code is 200 actual %s"
                             % response.status_code)

    def request_post(self, api, header, payload):
        """Executes POST request to create a resource"""

        creds = RequestScenario.get_credentials()
        url = creds.get('cli', 'api_url') + '/' + api
        response = requests.request("POST",
                                    url,
                                    data=jsonutils.dumps(payload),
                                    headers=header)
        if response.status_code == 201:
            LOG.debug("Resource creation successful for: '%s'" % api)
            return response.status_code
        elif response.status_code == 401:
            return response.status_code
        else:
            LOG.error("Error for the operation is: %s" % response.text)
            raise ValueError("Expected HTTP request code is 201 actual %s"
                             % response.status_code)

    def request_put(self, api, header, payload):
        """Executes PUT request to update an existing resource"""

        creds = RequestScenario.get_credentials()
        url = creds.get('cli', 'api_url') + '/' + api
        response = requests.request("PUT",
                                    url,
                                    data=jsonutils.dumps(payload),
                                    headers=header)
        if response.status_code == 204:
            LOG.debug("Resource '%s' updated successfully" % api)
        elif response.status_code == 401:
            return response.status_code
        else:
            LOG.error("Error for the operation is: %s" % response.text)
            raise ValueError("Expected HTTP request code is 204 actual %s"
                             % response.status_code)

    def request_delete(self, api, header):
        """Executes DELETE request to delete a resource"""

        creds = RequestScenario.get_credentials()
        url = creds.get('cli', 'api_url') + '/' + api
        response = requests.request("DELETE", url, headers=header)
        if response.status_code == 204:
            LOG.debug("Resource delete successful for: '%s'" % api)
        elif response.status_code == 401:
            return response.status_code
        else:
            LOG.error("Error for the operation is: %s" % response.text)
            raise ValueError("Expected HTTP request code is 204 actual %s"
                             % response.status_code)



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import types

import six

from neutron_lib import constants as n_const

from neutron.common import exceptions as n_exc


def check_update_port(orig, new):
    device_owner = new['device_owner']
    # REVISIT(yamamoto): The empty device_owner is allowed here because
    # it's assumed by many of unit tests and tempest.
    if (new['fixed_ips'] != orig['fixed_ips'] and
        not (device_owner.startswith(n_const.DEVICE_OWNER_COMPUTE_PREFIX) or
             device_owner.startswith(n_const.DEVICE_OWNER_ROUTER_INTF) or
             device_owner == '')):
        raise n_exc.UnsupportedPortDeviceOwner(op='fixed_ips update',
                                               port_id=new['id'],
                                               device_owner=device_owner)


def unboundmethod(func, cls):
    if six.PY3:
        # python 3.x doesn't have unbound methods
        func.__qualname__ = cls.__qualname__ + '.' + func.__name__  # PEP 3155
        return func
    else:  # python 2.x
        return types.MethodType(func, None, cls)



# Copyright (C) 2016 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.exceptions import firewall_v1 as fw_exc
from oslo_db import exception as db_exc
from oslo_log import helpers as log_helpers
from oslo_utils import uuidutils
from sqlalchemy.orm import exc

from neutron.db import common_db_mixin

from midonet.neutron.db import logging_resource_model as model
from midonet.neutron.extensions import logging_resource as log_res_ext


class LoggingResourceDbMixin(log_res_ext.LoggingResourcePluginBase,
                             common_db_mixin.CommonDbMixin):
    """Mixin class to add logging resource."""

    __native_bulk_support = False

    @log_helpers.log_method_call
    def create_logging_resource(self, context, logging_resource):
        """Create a logging_resource"""
        log_res = logging_resource['logging_resource']
        with context.session.begin(subtransactions=True):
            log_res_db = model.LoggingResource(
                id=uuidutils.generate_uuid(),
                name=log_res['name'],
                description=log_res['description'],
                tenant_id=log_res['tenant_id'],
                enabled=log_res['enabled'])
            context.session.add(log_res_db)

        return self._make_logging_resource_dict(log_res_db)

    @log_helpers.log_method_call
    def get_logging_resource(self, context, id, fields=None):
        log_res_db = self._get_logging_resource(context, id)
        return self._make_logging_resource_dict(log_res_db, fields)

    @log_helpers.log_method_call
    def get_logging_resources(self, context, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'logging_resource',
                                          limit, marker)

        return self._get_collection(
            context,
            model.LoggingResource,
            self._make_logging_resource_dict,
            filters=filters, fields=fields,
            sorts=sorts,
            limit=limit, marker_obj=marker_obj,
            page_reverse=page_reverse)

    @log_helpers.log_method_call
    def update_logging_resource(self, context, id, logging_resource):
        log_res = logging_resource['logging_resource']
        with context.session.begin(subtransactions=True):
            log_res_db = self._get_logging_resource(context, id)
            log_res_db.update(log_res)

        return self._make_logging_resource_dict(log_res_db)

    @log_helpers.log_method_call
    def delete_logging_resource(self, context, id):
        log_res_db = self._get_logging_resource(context, id)
        with context.session.begin(subtransactions=True):
            context.session.delete(log_res_db)

    @log_helpers.log_method_call
    def delete_logging_resource_firewall_log(self, context, id,
                                             logging_resource_id):
        f_log_db = self._get_firewall_log(context, id)
        with context.session.begin(subtransactions=True):
            context.session.delete(f_log_db)

    @log_helpers.log_method_call
    def create_logging_resource_firewall_log(
            self, context, firewall_log, logging_resource_id):
        f_log = firewall_log['firewall_log']
        try:
            with context.session.begin(subtransactions=True):
                f_log_db = model.FirewallLog(
                    id=uuidutils.generate_uuid(),
                    logging_resource_id=logging_resource_id,
                    tenant_id=f_log['tenant_id'],
                    description=f_log['description'],
                    fw_event=f_log['fw_event'],
                    firewall_id=f_log['firewall_id'])
                context.session.add(f_log_db)
        except db_exc.DBReferenceError:
            raise fw_exc.FirewallNotFound(firewall_id=f_log['firewall_id'])

        return self._make_firewall_log_dict(f_log_db)

    @log_helpers.log_method_call
    def get_logging_resource_firewall_log(self, context, id,
                                          logging_resource_id, fields=None):
        f_log_db = self._get_firewall_log(context, id)
        return self._make_firewall_log_dict(f_log_db, fields)

    @log_helpers.log_method_call
    def get_logging_resource_firewall_logs(self, context, logging_resource_id,
                                           filters=None, fields=None,
                                           sorts=None, limit=None,
                                           marker=None, page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'firewall_log', limit,
                                          marker)
        return self._get_collection(context,
                                    model.FirewallLog,
                                    self._make_firewall_log_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts,
                                    limit=limit, marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    @log_helpers.log_method_call
    def get_firewall_logs(self, context, filters=None, fields=None,
                          sorts=None, limit=None, marker=None,
                          page_reverse=False):
        # This method is only for quota driver.
        return self._get_fw_logs_from_tenant(context, context.tenant_id)

    @log_helpers.log_method_call
    def update_logging_resource_firewall_log(
            self, context, id, logging_resource_id, firewall_log):
        f_log = firewall_log['firewall_log']
        with context.session.begin(subtransactions=True):
            f_log_db = self._get_firewall_log(context, id)
            f_log_db.update(f_log)

        return self._make_firewall_log_dict(f_log_db)

    def _get_logging_resource(self, context, id):
        try:
            query = self._model_query(context, model.LoggingResource)
            log_res_db = query.filter(model.LoggingResource.id == id).one()

        except exc.NoResultFound:
            raise log_res_ext.LoggingResourceNotFound(id=id)

        return log_res_db

    def _get_fw_logs_from_tenant(self, context, tenant_id):
        query = self._model_query(context, model.FirewallLog)
        return query.filter(model.FirewallLog.tenant_id == tenant_id).all()

    def _logging_resource_has_logs(self, context, log_res_id):
        query = self._model_query(context, model.FirewallLog)
        return bool(query.filter(
            model.FirewallLog.logging_resource_id == log_res_id).all())

    def _get_firewall_log(self, context, id):
        try:
            query = self._model_query(context, model.FirewallLog)
            f_log_db = query.filter(model.FirewallLog.id == id).one()

        except exc.NoResultFound:
            raise log_res_ext.FirewallLogNotFound(id=id)

        return f_log_db

    def _make_logging_resource_dict(self, log_res_db, fields=None):
        res = {'id': log_res_db['id'],
               'name': log_res_db['name'],
               'description': log_res_db['description'],
               'tenant_id': log_res_db['tenant_id'],
               'enabled': log_res_db['enabled']}

        # extend parameter for specific resource types
        res.update(self._extend_resource_specific_loggings(log_res_db))

        return self._fields(res, fields)

    def _extend_resource_specific_loggings(self, log_res_db, fields=None):
        # Currently, only firewall log is returned.
        return {
            'firewall_logs': [
                self._make_firewall_log_dict(f_log, fields)
                for f_log in log_res_db.firewall_logs
                if log_res_db.firewall_logs
            ]
        }

    def _make_firewall_log_dict(self, f_log_db, fields=None):
        res = {'id': f_log_db['id'],
               'tenant_id': f_log_db['tenant_id'],
               'description': f_log_db['description'],
               'logging_resource_id': f_log_db['logging_resource_id'],
               'fw_event': f_log_db['fw_event'],
               'firewall_id': f_log_db['firewall_id']}
        return self._fields(res, fields)



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron.ipam import utils as ipam_utils
from neutron_dynamic_routing.db import bgp_db

LOG = logging.getLogger(__name__)


class MidonetBgpDbMixin(bgp_db.BgpDbMixin):

    """Access methods to figure out advertised route by bgp speaker."""

    @log_helpers.log_method_call
    def get_routes_from_attached_networks(self, context,
                                          bgp_speaker_id, rt_id):
        """This method gets routes for networks attached to this router.

        * 'nexthop' is the IP address of the ports on this router
          that are on the same subnet as the peer IP addresses of
          the provided bgp speaker.
        * 'dest_networks' are all the networks attached to the router
          excluding gateway and uplink networks.
        """
        nexthops, dest_networks, route_info = [], [], []
        peers = self.get_bgp_peers_by_bgp_speaker(context, bgp_speaker_id)
        peer_ips = [peer['peer_ip'] for peer in peers]
        LOG.debug("peer_ips %s associated with bgp_speaker %s",
                  peer_ips, bgp_speaker_id)
        core_plugin = directory.get_plugin()
        ports = core_plugin.get_ports(context, filters={'device_id': [rt_id]})
        for port in ports:
            subnet_id = port['fixed_ips'][0]['subnet_id']
            cidr = core_plugin.get_subnet(
                context, subnet_id, fields=['cidr'])
            if self._extract_valid_peer_ips(cidr['cidr'], peer_ips):
                nexthops.append(port['fixed_ips'][0]['ip_address'])
            dest_networks.append(cidr['cidr'])
        for nexthop in nexthops:
            route_info += self._make_route_info(nexthop, dest_networks)
        LOG.debug("advertised routes from attached networks: %s", route_info)
        return route_info

    @log_helpers.log_method_call
    def get_routes_from_extra_routes(self, context, rt_id, nexthops):
        """This method gets routes for extra routes defined on this router.

        * 'nexthop' is the IP address of the ports on this router
          that are on the same subnet as the peer IP addresses of
          the provided bgp speaker.
        * 'dests' are all the networks defined as extra routes
          on the router.
        """
        changed_extra = []
        l3plugin = directory.get_plugin(constants.L3)
        extra = l3plugin.get_router(context, rt_id,
                                    fields=['routes'])['routes']
        dest_networks = [extra_route['destination'] for extra_route in extra]
        for nexthop in nexthops:
            changed_extra += self._change_nexthop_to_router_ip(nexthop,
                                                               dest_networks)
        LOG.debug("advertised routes from extra routes: %s", changed_extra)
        return changed_extra

    def _change_nexthop_to_router_ip(self, nexthop, dests):
        return [(dest, nexthop) for dest in dests]

    def _make_route_info(self, nexthop, dest_networks):
        return [(network, nexthop) for network in dest_networks
                if not ipam_utils.check_subnet_ip(network, nexthop)]

    def _extract_valid_peer_ips(self, cidr, peer_ips):
        return [(peer_ip) for peer_ip in peer_ips
                if ipam_utils.check_subnet_ip(cidr, peer_ip)]



# Copyright 2016 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Remove agent membership

Revision ID: 1612b5389e6e
Revises: 27e6e3451f22
Create Date: 2016-12-09 00:27:26.878502

"""

from alembic import op

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = '1612b5389e6e'
down_revision = '27e6e3451f22'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.OCATA, migration.PIKE]


def upgrade():
    op.drop_table('midonet_agent_memberships')



# Copyright 2015 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""add gateway device management

Revision ID: cfe0dea89aa
Revises: 4f3b347ea1c2
Create Date: 2015-12-21 11:06:46.155138

"""

# revision identifiers, used by Alembic.
revision = 'cfe0dea89aa'
down_revision = '4f3b347ea1c2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('midonet_gateway_devices',
                    sa.Column('id', sa.String(length=36), primary_key=True),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('type', sa.String(length=255), nullable=False),
                    sa.Column('tenant_id', sa.String(length=255),
                              nullable=True))

    op.create_table('midonet_gateway_hw_vtep_devices',
                    sa.Column('device_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('management_ip', sa.String(length=64),
                              nullable=False, unique=True),
                    sa.Column('management_port', sa.Integer(),
                              nullable=False),
                    sa.Column('management_protocol', sa.String(length=255),
                              nullable=False),
                    sa.ForeignKeyConstraint(['device_id'],
                                            ['midonet_gateway_devices.id'],
                                            ondelete='CASCADE'))

    op.create_table('midonet_gateway_overlay_router_devices',
                    sa.Column('device_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('resource_id', sa.String(length=36),
                              nullable=False),
                    sa.ForeignKeyConstraint(['device_id'],
                                            ['midonet_gateway_devices.id'],
                                            ondelete='CASCADE'))

    op.create_table('midonet_gateway_tunnel_ips',
                    sa.Column('device_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('tunnel_ip', sa.String(length=64),
                              nullable=False, unique=True),
                    sa.ForeignKeyConstraint(['device_id'],
                                            ['midonet_gateway_devices.id'],
                                            ondelete='CASCADE'))

    op.create_table('midonet_gateway_remote_mac_tables',
                    sa.Column('id', sa.String(length=36),
                              primary_key=True),
                    sa.Column('device_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('mac_address', sa.String(length=32),
                              nullable=False, unique=True),
                    sa.Column('vtep_address', sa.String(length=64),
                              nullable=False),
                    sa.Column('segmentation_id', sa.Integer(),
                              nullable=True),
                    sa.ForeignKeyConstraint(['device_id'],
                                            ['midonet_gateway_devices.id'],
                                            ondelete='CASCADE'))



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources

from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

LOG = logging.getLogger(__name__)


@registry.has_registry_receivers
class MidonetSecurityGroupsHandler(object):

    def __init__(self, client):
        self.client = client

    # TODO(yamamoto): Implement PRECOMMIT hooks for task-based api
    # (bug 1694699)
    @registry.receives(resources.SECURITY_GROUP, [events.AFTER_CREATE])
    @log_helpers.log_method_call
    def create_security_group(self, resource, event, trigger, **kwargs):
        sg = kwargs.get('security_group')
        try:
            self.client.create_security_group_postcommit(sg)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create a security group %(sg_id)s "
                          "in Midonet: %(err)s",
                          {"sg_id": sg["id"], "err": ex})
                try:
                    self.client.delete_security_group_postcommit(sg["id"])
                except Exception:
                    LOG.exception("Failed to delete security group %s",
                                  sg['id'])

    @registry.receives(resources.SECURITY_GROUP, [events.AFTER_UPDATE])
    @log_helpers.log_method_call
    def update_security_group(self, resource, event, trigger, **kwargs):
        pass

    @registry.receives(resources.SECURITY_GROUP, [events.AFTER_DELETE])
    @log_helpers.log_method_call
    def delete_security_group(self, resource, event, trigger, **kwargs):
        sg_id = kwargs.get('security_group_id')
        try:
            self.client.delete_security_group_postcommit(sg_id)
        except Exception as ex:
            LOG.error("Failed to a delete security group %(sg_id)s "
                      "in Midonet: %(err)s",
                      {"sg_id": sg_id, "err": ex})

    @registry.receives(resources.SECURITY_GROUP_RULE, [events.AFTER_CREATE])
    @log_helpers.log_method_call
    def create_security_group_rule(self, resource, event, trigger, **kwargs):
        sgr = kwargs.get('security_group_rule')
        try:
            self.client.create_security_group_rule_postcommit(sgr)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create a security group rule "
                          "%(sgr_id)s in Midonet: %(err)s",
                          {"sgr_id": sgr["id"], "err": ex})
                try:
                    self.client.delete_security_group_rule_postcommit(
                        sgr["id"])
                except Exception:
                    LOG.exception("Failed to delete security group  rule %s",
                                  sgr['id'])

    @registry.receives(resources.SECURITY_GROUP_RULE, [events.AFTER_DELETE])
    @log_helpers.log_method_call
    def delete_security_group_rule(self, resource, event, trigger, **kwargs):
        sgr_id = kwargs.get('security_group_rule_id')
        try:
            self.client.delete_security_group_rule_postcommit(sgr_id)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to delete a security group %(sgr_id)s "
                          "in Midonet: %(err)s",
                          {"sgr_id": sgr_id, "err": ex})



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

from neutron_taas.services.taas import service_drivers as taas_service_drivers

from midonet.neutron.client import base as c_base
from midonet.neutron.common import config  # noqa

LOG = logging.getLogger(__name__)


class MidonetTaasDriver(taas_service_drivers.TaasBaseDriver):
    """Midonet Taas Service Driver class"""

    def __init__(self, service_plugin):
        LOG.debug("Loading MidonetTaasDriver.")
        self.client = c_base.load_client(cfg.CONF.MIDONET)
        super(MidonetTaasDriver, self).__init__(service_plugin)

    @log_helpers.log_method_call
    def create_tap_service_precommit(self, context):
        # TODO(yamamoto): Call self.client for task-based api
        pass

    @log_helpers.log_method_call
    def create_tap_service_postcommit(self, context):
        try:
            ts = context.tap_service
            self.client.create_tap_service(context._plugin_context, ts)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create tap service %(tap_service)s "
                          "in Midonet: %(err)s",
                          {"tap_service": ts, "err": ex})

    @log_helpers.log_method_call
    def delete_tap_service_precommit(self, context):
        # TODO(yamamoto): Call self.client for task-based api
        pass

    @log_helpers.log_method_call
    def delete_tap_service_postcommit(self, context):
        try:
            ts = context.tap_service
            self.client.delete_tap_service(context._plugin_context, ts['id'])
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to delete a tap service"
                          "%(tap_service)s in Midonet: %(err)s",
                          {"tap_service": ts['id'], "err": ex})

    @log_helpers.log_method_call
    def create_tap_flow_precommit(self, context):
        # TODO(yamamoto): Call self.client for task-based api
        pass

    @log_helpers.log_method_call
    def create_tap_flow_postcommit(self, context):
        try:
            tf = context.tap_flow
            self.client.create_tap_flow(context._plugin_context, tf)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create tap flow %(tap_flow)s "
                          "in Midonet: %(err)s",
                          {"tap_flow": tf, "err": ex})

    @log_helpers.log_method_call
    def delete_tap_flow_precommit(self, context):
        # TODO(yamamoto): Call self.client for task-based api
        pass

    @log_helpers.log_method_call
    def delete_tap_flow_postcommit(self, context):
        try:
            tf = context.tap_flow
            self.client.delete_tap_flow(context._plugin_context, tf['id'])
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to delete a tap flow"
                          "%(tap_flow)s in Midonet: %(err)s",
                          {"tap_flow": tf['id'], "err": ex})



# Copyright (c) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from neutron_lib import constants as n_const

from neutron.agent.linux import interface as n_interface
from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils

LOG = logging.getLogger(__name__)


class MidonetInterfaceDriver(n_interface.LinuxInterfaceDriver):

    def plug_new(self, network_id, port_id, device_name, mac_address,
                 bridge=None, namespace=None, prefix=None, mtu=None):
        """Plug an interface.

        This method is called by the Dhcp agent or by the L3 agent
        when a new network is created
        """
        ip = ip_lib.IPWrapper()
        tap_name = device_name.replace(prefix or n_const.TAP_DEVICE_PREFIX,
                                       n_const.TAP_DEVICE_PREFIX)

        # Create ns_dev in a namespace if one is configured.
        root_dev, ns_dev = ip.add_veth(tap_name, device_name,
                                       namespace2=namespace)
        root_dev.disable_ipv6()
        ns_dev.link.set_address(mac_address)

        if mtu:
            ns_dev.link.set_mtu(mtu)
            root_dev.link.set_mtu(mtu)
        else:
            LOG.warning("No MTU configured for port %s", port_id)

        ns_dev.link.set_up()
        root_dev.link.set_up()

        cmd = ['mm-ctl', '--bind-port', port_id, device_name]
        utils.execute(cmd, run_as_root=True)  # nosec

    def unplug(self, device_name, bridge=None, namespace=None, prefix=None):
        # the port will be deleted by the dhcp agent that will call the plugin
        device = ip_lib.IPDevice(device_name, namespace=namespace)
        try:
            device.link.delete()
        except RuntimeError:
            LOG.error("Failed unplugging interface '%s'", device_name)
        LOG.debug("Unplugged interface '%s'", device_name)

        ip_lib.IPWrapper(namespace=namespace).garbage_collect_namespace()



# Copyright (C) 2012 Midokura Japan K.K.
# Copyright (C) 2013 Midokura PTE LTD
# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Import all data models
from networking_l2gw.db.l2gateway import l2gateway_models  # noqa

from midonet.neutron.client import base as cli_base
# Import all data models
from midonet.neutron.common import config  # noqa
from midonet.neutron.db.migration.models import head  # noqa


TEST_MN_CLIENT = ('midonet.neutron.tests.unit.test_midonet_plugin.'
                  'NoopMidonetClient')


class NoopMidonetClient(cli_base.MidonetClientBase):
    """Dummy midonet client used for the unit tests"""
    pass



# Copyright (C) 2016 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from neutronclient.common import extension
from neutronclient.neutron import v2_0 as loggingV20

from midonet.neutron._i18n import _


def add_common_paramaters_to_arguments(parser):
    parser.add_argument(
        '--enabled',
        metavar='<True | False>',
        choices=['True', 'False'],
        help=_('Enable/disable for the logging. Defaults to False.'))
    parser.add_argument(
        '--description',
        dest='description',
        help=_('Description of logging.'))


def _format_firewall_logs(logging_resource):
    try:
        return '\n'.join([jsonutils.dumps(fw_log) for fw_log
                          in logging_resource['firewall_logs']])
    except (TypeError, KeyError):
        return ''


class LoggingResource(extension.NeutronClientExtension):
    resource = 'logging_resource'
    resource_plural = 'logging_resources'
    path = 'logging/logging_resources'
    object_path = '/%s' % path
    resource_path = '/%s/%%s' % path
    versions = ['2.0']


class LoggingResourceCreate(extension.ClientExtensionCreate, LoggingResource):
    """Create Logging Resource information."""

    shell_command = 'logging-create'

    def add_known_arguments(self, parser):
        parser.add_argument(
            'name', metavar='NAME',
            help=_('User defined logging name.'))
        add_common_paramaters_to_arguments(parser)

        return parser

    def args2body(self, args):
        body = {}
        attributes = ['name', 'description', 'tenant_id', 'enabled']
        loggingV20.update_dict(args, body, attributes)

        return {'logging_resource': body}


class LoggingResourceList(extension.ClientExtensionList, LoggingResource):
    """List Logging Resources."""

    shell_command = 'logging-list'
    list_columns = ['id', 'name', 'enabled']
    pagination_support = True
    sorting_support = True

    _formatters = {'firewall_logs': _format_firewall_logs, }


class LoggingResourceShow(extension.ClientExtensionShow, LoggingResource):
    """Show information of a given Logging Resource."""

    shell_command = 'logging-show'


class LoggingResourceDelete(extension.ClientExtensionDelete, LoggingResource):
    """Delete a given Logging Resource."""

    shell_command = 'logging-delete'


class LoggingResourceUpdate(extension.ClientExtensionUpdate, LoggingResource):
    """Update a given Logging Resource."""

    shell_command = 'logging-update'

    def add_known_arguments(self, parser):
        parser.add_argument(
            '--name', dest='name',
            help=_('User defined logging name.'))
        add_common_paramaters_to_arguments(parser)

    def args2body(self, args):
        body = {}
        attributes = ['name', 'description', 'enabled']
        loggingV20.update_dict(args, body, attributes)

        return {'logging_resource': body}



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from neutronclient import shell

from midonet.neutron.tests.unit.neutronclient_ext import test_cli20
from midonet.neutronclient.gateway_device_extension import _gateway_device


class CLITestV20GatewayDeviceJSON(test_cli20.CLIExtTestV20Base):

    def setUp(self):
        gw_device = ("gateway_device", _gateway_device)
        self._mock_load_extensions(gw_device)
        super(CLITestV20GatewayDeviceJSON,
              self).setUp(plurals={'gateway_devices': 'gateway_device'})
        self.register_non_admin_status_resource('gateway_device')

    def test_gateway_device_cmd_loaded(self):
        neutron_shell = shell.NeutronShell('2.0')
        gw_device_cmd = {'gateway-device-list':
                         _gateway_device.GatewayDeviceList,
                         'gateway-device-create':
                         _gateway_device.GatewayDeviceCreate,
                         'gateway-device-update':
                         _gateway_device.GatewayDeviceUpdate,
                         'gateway-device-delete':
                         _gateway_device.GatewayDeviceDelete,
                         'gateway-device-show':
                         _gateway_device.GatewayDeviceShow
                         }
        for cmd_name, cmd_class in gw_device_cmd.items():
            found = neutron_shell.command_manager.find_command([cmd_name])
            self.assertEqual(cmd_class, found[0])

    def _create_gateway_device(self, name, args,
                               position_names, position_values):
        resource = 'gateway_device'
        cmd = _gateway_device.GatewayDeviceCreate(
            test_cli20.MyApp(sys.stdout), None)
        self._test_create_resource(resource, cmd, name, 'myid',
                                   args, position_names, position_values)

    def _update_gateway_device(self, args, values):
        resource = 'gateway_device'
        cmd = _gateway_device.GatewayDeviceUpdate(
            test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid', args, values)

    def test_create_gateway_device_for_hw_vtep_mandatory_params(self):
        name = 'hw-vtep-mandatory'
        gw_type = 'hw_vtep'
        mng_ip = '10.0.0.100'
        mng_port = '22'
        mng_protocol = 'ovsdb'
        args = ['--type', gw_type,
                '--management-ip', mng_ip,
                '--management-port', mng_port,
                '--management-protocol', mng_protocol]
        position_names = ['type', 'management_ip',
                          'management_port', 'management_protocol']
        position_values = [gw_type, mng_ip, mng_port, mng_protocol]
        self._create_gateway_device(name, args,
                                    position_names, position_values)

    def test_create_gateway_device_for_hw_vtep_with_optional_params(self):
        name = 'hw-vtep-optional'
        tenant_id = 'my_tenant'
        gw_type = 'hw_vtep'
        mng_ip = '10.0.0.100'
        mng_port = '22'
        mng_protocol = 'ovsdb'
        tunnel_ip = '200.200.200.4'
        args = ['--tenant-id', tenant_id,
                '--type', gw_type,
                '--management-ip', mng_ip,
                '--management-port', mng_port,
                '--management-protocol', mng_protocol,
                '--name', name,
                '--tunnel-ip', tunnel_ip]
        position_names = ['type', 'management_ip', 'management_port',
                          'management_protocol', 'tenant_id',
                          'name', 'tunnel_ips']
        position_values = [gw_type, mng_ip, mng_port, mng_protocol,
                           tenant_id, name, [tunnel_ip]]
        self._create_gateway_device(name, args,
                                    position_names, position_values)

    def test_create_gateway_device_for_router_vtep_with_mandatory_params(self):
        name = 'router-vtep-mandatory'
        gw_type = 'router_vtep'
        resource_id = 'my_router_id'
        args = ['--type', gw_type, '--resource-id', resource_id]
        position_names = ['type', 'resource_id']
        position_values = [gw_type, resource_id]
        self._create_gateway_device(name, args,
                                    position_names, position_values)

    def test_create_gateway_device_for_router_vtep_with_optional_params(self):
        name = 'router-vtep-optional'
        tenant_id = 'my_tenant'
        gw_type = 'router_vtep'
        resource_id = 'my_router_id'
        tunnel_ip = '200.200.200.4'
        args = ['--tenant-id', tenant_id,
                '--type', gw_type,
                '--resource-id', resource_id,
                '--name', name,
                '--tunnel-ip', tunnel_ip]
        position_names = ['type', 'resource_id',
                          'tenant_id', 'name', 'tunnel_ips']
        position_values = [gw_type, resource_id,
                           tenant_id, name, [tunnel_ip]]
        self._create_gateway_device(name, args,
                                    position_names, position_values)

    def test_create_gateway_device_for_network_vlan_with_mandatory_params(
            self):
        name = 'network_vlan-mandatory'
        gw_type = 'network_vlan'
        resource_id = 'my_network_id'
        args = ['--type', gw_type, '--resource-id', resource_id]
        position_names = ['type', 'resource_id']
        position_values = [gw_type, resource_id]
        self._create_gateway_device(name, args,
                                    position_names, position_values)

    def test_create_gateway_device_for_network_vlan_with_optional_params(self):
        name = 'network_vlan-optional'
        tenant_id = 'my_tenant'
        gw_type = 'network_vlan'
        resource_id = 'my_network_id'
        args = ['--tenant-id', tenant_id,
                '--type', gw_type,
                '--resource-id', resource_id,
                '--name', name]
        position_names = ['type', 'resource_id',
                          'tenant_id', 'name']
        position_values = [gw_type, resource_id,
                           tenant_id, name]
        self._create_gateway_device(name, args,
                                    position_names, position_values)

    def test_update_gateway_device_with_name(self):
        args = ['myid', '--name', 'name_updated']
        values = {'name': 'name_updated'}
        self._update_gateway_device(args, values)

    def test_update_gateway_device_with_tunnel_ip(self):
        args = ['myid', '--tunnel-ip', '200.200.200.4']
        values = {'tunnel_ips': ['200.200.200.4']}
        self._update_gateway_device(args, values)

    def test_update_gateway_device_with_tunnel_ips(self):
        args = ['myid', '--tunnel-ip', '200.200.200.4',
                '--tunnel-ip', '200.200.200.10']
        values = {'tunnel_ips': ['200.200.200.4',
                                 '200.200.200.10']}
        self._update_gateway_device(args, values)

    def test_delete_gateway_device(self):
        resource = 'gateway_device'
        cmd = _gateway_device.GatewayDeviceDelete(
            test_cli20.MyApp(sys.stdout), None)
        my_id = 'my-id'
        args = [my_id]
        self._test_delete_resource(resource, cmd, my_id, args)

    def test_list_gateway_devices(self):
        resources = 'gateway_devices'
        cmd = _gateway_device.GatewayDeviceList(
            test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd)

    def test_list_gateway_devices_with_pagination(self):
        resources = 'gateway_devices'
        cmd = _gateway_device.GatewayDeviceList(
            test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources_with_pagination(resources, cmd)

    def test_list_gateway_device_with_remote_mac_entries(self):
        resources = 'gateway_devices'
        cmd = _gateway_device.GatewayDeviceList(
            test_cli20.MyApp(sys.stdout), None)
        rme = [
            {"segmentation_id": 100,
             "vtep_address": "192.168.100.1",
             "id": "remote_mac_entry_id1",
             "mac_address": "fa:16:3e:db:79:80"},
            {"segmentation_id": 100,
             "vtep_address": "192.168.100.50",
             "id": "remote_mac_entry_id1",
             "mac_address": "fa:16:3e:df:79:80"},
        ]
        response = {'gateway_devices': [{"id": 'myid',
                                         "name": 'gw_device',
                                         "type": "router_vtep",
                                         "resource_id": "router_id",
                                         "tunnel_ips": [],
                                         "remote_mac_entries": rme}]}
        args = ['-c', 'id', '-c', 'remote_mac_entries']
        self._test_list_columns(cmd, resources, response, args)

    def test_show_gateway_device(self):
        resource = 'gateway_device'
        cmd = _gateway_device.GatewayDeviceShow(
            test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id, args,
                                 ['id', 'name'])



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket
import uuid

from midonetclient import topology  # noqa
from midonetclient.topology import hosts
from midonetclient.topology import pool_members

from midonet.neutron.common import exceptions as mexc


MIDO_AGENT_NAME = 'Midonet Agent'
MIDO_BINARY = 'midolman'


def midonet_host_to_neutron_agent(host):
    return {'admin_state_up': True,
            'agent_type': MIDO_AGENT_NAME,
            'alive': True,
            'binary': MIDO_BINARY,
            'configurations': {},
            'host': None,
            'id': host.get('id')}


def invoke_cluster_rpc(cluster_ip, cluster_port, call):
    try:
        sock = socket.create_connection((cluster_ip, cluster_port))
        req_uuid = uuid.uuid4()
        topology.handshake(sock, uuid.uuid4(), req_uuid)
        ret = list(call(sock))
        topology.bye(sock, req_uuid)
        return ret
    except socket.error:
        raise mexc.ClusterConnectionError()


def get_all_midonet_hosts(cluster_ip, cluster_port):
    return invoke_cluster_rpc(cluster_ip, cluster_port, hosts.get_all_dict)


def get_all_midonet_pool_members(cluster_ip, cluster_port):
    return invoke_cluster_rpc(cluster_ip, cluster_port,
                              pool_members.get_all_dict)



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six

from oslo_config import cfg

from neutron_lib.api.definitions import logging_resource
from neutron_lib.api import extensions as api_extensions
from neutron_lib import exceptions as nexception
from neutron_lib.plugins import directory

from neutron.api import extensions
from neutron.api.v2 import base
from neutron.api.v2 import resource_helper
from neutron.quota import resource_registry

from midonet.neutron._i18n import _
from midonet.neutron.common import constants


class LoggingResourceNotFound(nexception.NotFound):
    message = _("Logging resource %(id)s does not exist")


class FirewallLogNotFound(nexception.NotFound):
    message = _("Firewall log %(id)s does not exist")


class ResourceInUseByLoggingResource(nexception.InUse):
    message = _("%(resource_name)s %(resource_id)s %(reason)s")

    def __init__(self, **kwargs):
        if 'reason' not in kwargs:
            kwargs['reason'] = "is in use by logging resource"
        super(ResourceInUseByLoggingResource, self).__init__(**kwargs)


FW_EVENT_ACCEPT = logging_resource.FW_EVENT_ACCEPT
FW_EVENT_DROP = logging_resource.FW_EVENT_DROP
FW_EVENT_ALL = logging_resource.FW_EVENT_ALL
FW_EVENTS = logging_resource.FW_EVENTS

# A tenant may want to create firewall logs for all firewalls in tenant.
# Set default quotas to align with default quota_firewall of 10.

firewall_log_quota_opts = [
    cfg.IntOpt('quota_firewall_log',
               default=10,
               help=_('Number of firewall logs allowed per tenant. '
                      'A negative value means unlimited.'))
]
cfg.CONF.register_opts(firewall_log_quota_opts, 'QUOTAS')


class Logging_resource(api_extensions.APIExtensionDescriptor):
    """Logging resource extension."""

    api_definition = logging_resource

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""

        plural_mappings = resource_helper.build_plural_mappings(
            {}, logging_resource.RESOURCE_ATTRIBUTE_MAP)

        resources = resource_helper.build_resource_info(
            plural_mappings,
            logging_resource.RESOURCE_ATTRIBUTE_MAP,
            constants.LOGGING_RESOURCE)
        plugin = directory.get_plugin(constants.LOGGING_RESOURCE)

        for collection_name in logging_resource.SUB_RESOURCE_ATTRIBUTE_MAP:
            resource_name = collection_name[:-1]
            resource_registry.register_resource_by_name(resource_name)
            parent = logging_resource.SUB_RESOURCE_ATTRIBUTE_MAP[
                collection_name].get('parent')
            params = logging_resource.SUB_RESOURCE_ATTRIBUTE_MAP[
                collection_name].get('parameters')

            controller = base.create_resource(collection_name, resource_name,
                                              plugin, params,
                                              allow_bulk=True,
                                              parent=parent)

            resource = extensions.ResourceExtension(
                collection_name,
                controller, parent,
                path_prefix=logging_resource.API_PREFIX,
                attr_map=params)
            resources.append(resource)

        return resources


@six.add_metaclass(abc.ABCMeta)
class LoggingResourcePluginBase(object):

    path_prefix = logging_resource.API_PREFIX

    @abc.abstractmethod
    def create_logging_resource(self, context, logging_resource):
        pass

    @abc.abstractmethod
    def update_logging_resource(self, context, id, logging_resource):
        pass

    @abc.abstractmethod
    def delete_logging_resource(self, context, id):
        pass

    @abc.abstractmethod
    def get_logging_resources(self, context, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        pass

    @abc.abstractmethod
    def get_logging_resource(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def create_logging_resource_firewall_log(self, context,
                                             logging_resource_id,
                                             firewall_log):
        pass

    @abc.abstractmethod
    def update_logging_resource_firewall_log(
            self, context, id, logging_resource_id, firewall_log):
        pass

    @abc.abstractmethod
    def delete_logging_resource_firewall_log(self, context,
                                             id, logging_resource_id):
        pass

    @abc.abstractmethod
    def get_logging_resource_firewall_logs(self, context, logging_resource_id,
                                           filters=None, fields=None,
                                           sorts=None, limit=None,
                                           marker=None, page_reverse=False):
        pass

    @abc.abstractmethod
    def get_logging_resource_firewall_log(self, context, id,
                                          logging_resource_id, fields=None):
        pass



# Copyright (C) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
import six

from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway import service_drivers

from midonet.neutron.client import base as c_base
from midonet.neutron.common import config  # noqa

LOG = logging.getLogger(__name__)


# TODO(yamamoto): Update for the new driver API (See bug 1577339)
# and override precommit methods for task-based api.


@six.add_metaclass(abc.ABCMeta)
class MidonetL2gwDriver(service_drivers.L2gwDriver):
    """L2GW MidoNet Service Driver class."""

    def __init__(self, service_plugin, validator=None):
        super(MidonetL2gwDriver, self).__init__(service_plugin,
                                                validator=validator)
        self.service_plugin = service_plugin
        self.client = c_base.load_client(cfg.CONF.MIDONET)

    @property
    def service_type(self):
        return 'l2gw'

    def add_port_mac(self, context, port_dict):
        raise NotImplementedError()

    def delete_port_mac(self, context, port):
        raise NotImplementedError()

    def _validate_gw_connection(self, context, gw_connection):
        network_id = gw_connection.get(constants.NETWORK_ID)
        self.service_plugin._get_network(context, network_id)

    def _make_gateway_dict(self, context, gw_id):
        gw = self.service_plugin._get_l2_gateway(context, gw_id)
        return self.service_plugin._make_l2_gateway_dict(gw)

    def _make_gateway_conn_dict(self, context, gw_connection):
        gw_conn_dict = self.service_plugin._make_l2gw_connections_dict(
            gw_connection)
        gw_conn_dict['l2_gateway'] = self._make_gateway_dict(
            context, gw_connection["l2_gateway_id"])
        return gw_conn_dict

    def create_l2_gateway_connection(self, context, l2_gateway_connection):
        gw_connection = l2_gateway_connection.get(
            self.service_plugin.connection_resource)

        gw_conn_dict = self._make_gateway_conn_dict(context, gw_connection)
        self._validate_gw_connection(context, gw_conn_dict)
        if not gw_conn_dict[constants.SEG_ID]:
            seg_id = self.service_plugin._get_l2_gateway_seg_id(
                context, gw_conn_dict['l2_gateway_id'])
            gw_conn_dict[constants.SEG_ID] = seg_id
        try:
            self.client.create_l2_gateway_connection(context,
                                                     gw_conn_dict)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create L2GW Conn %(l2_gw_conn)s "
                          "in Midonet: %(err)s",
                          {"l2_gw_conn": gw_connection, "err": ex})

    def delete_l2_gateway_connection(self, context, l2_gateway_connection):
        try:
            self.client.delete_l2_gateway_connection(context,
                                                     l2_gateway_connection)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to delete a l2 Gateway Connection"
                          "%(l2_gw_conn_id)s in Midonet: %(err)s",
                          {"l2_gw_conn_id": l2_gateway_connection, "err": ex})
